import torch
import torch.nn as nn
import torch.nn.functional as F
import torchelie.nn as tnn
from torchelie.utils import experimental
from typing import Tuple


class FactoredPredictor(nn.Module):
    heads: nn.ModuleList

    def __init__(self, hid_ch: int, out_ch: int, n_pred: int) -> None:
        super(FactoredPredictor, self).__init__()
        self.heads = nn.ModuleList([
            nn.Sequential(nn.Linear(hid_ch + i, hid_ch + i),
                          nn.ReLU(inplace=True), nn.Linear(hid_ch + i, out_ch))
            for i in range(n_pred)
        ])

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        # NCHW -> NWHC
        x = x.transpose(1, -1)
        y = y.transpose(1, -1)
        # NWKHC
        out = torch.stack([
            self.heads[i](torch.cat([x, y[..., :i]], dim=-1))
            for i in range(len(self.heads))
        ], dim=2)
        # NCKHW
        return out.transpose(1, -1)

    def sample(self, x: torch.Tensor, temp: float) -> torch.Tensor:
        sampled = torch.empty(x.shape[0], 0, device=x.device)
        for i in range(len(self.heads)):
            logits = self.heads[i](torch.cat([x, self.normalize(sampled)],
                                             dim=1)) / temp
            samp = torch.distributions.Categorical(logits=logits,
                                                   validate_args=True).sample(
                                                       (1, ))
            samp = samp.t()
            sampled = torch.cat(
                [sampled, self.cls_to_val(samp.float())], dim=1)
        return sampled

    def normalize(self, x: torch.Tensor) -> torch.Tensor:
        return x

    def cls_to_val(self, cls: torch.Tensor) -> torch.Tensor:
        return cls.float()


class PixelPredictor(FactoredPredictor):
    def __init__(self, hid_ch: int, n_ch: int = 3):
        super(PixelPredictor, self).__init__(hid_ch, 256, n_ch)

    def normalize(self, x: torch.Tensor) -> torch.Tensor:
        return x * 2 - 1

    def cls_to_val(self, cls: torch.Tensor) -> torch.Tensor:
        return cls.float() / 255


class ResBlk(nn.Module):
    @experimental
    def __init__(self, in_ch: int, hid_ch: int, out_ch: int, ks: int,
                 sz: Tuple[int, int]) -> None:
        super(ResBlk, self).__init__()
        self.go = tnn.CondSeq(
            nn.BatchNorm2d(in_ch),
            nn.ReLU(inplace=False),
            tnn.Conv1x1(in_ch, hid_ch),
            nn.BatchNorm2d(hid_ch),
            nn.ReLU(inplace=True),
            tnn.TopLeftConv2d(hid_ch, hid_ch, ks, center=True, bias=sz),
            nn.BatchNorm2d(hid_ch),
            nn.ReLU(inplace=True),
            tnn.Conv1x1(hid_ch, out_ch),
        )

    def condition(self, z: torch.Tensor) -> None:
        self.go.condition(z)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.go(x)
        return x + out


# https://arxiv.org/pdf/1701.05517.pdf
class PixCNNBase(nn.Module):
    @experimental
    def __init__(self,
                 in_ch: int,
                 hid: int,
                 out_ch: int,
                 quant_lvls: int,
                 sz: Tuple[int, int],
                 n_layer: int = 3) -> None:
        super(PixCNNBase, self).__init__()
        self.sz = sz
        self.lin = tnn.CondSeq(
            tnn.TopLeftConv2d(in_ch, hid, 5, center=False, bias=sz),
            nn.ReLU(inplace=True))

        sz2 = sz[0] // 2, sz[1] // 2
        sz4 = sz[0] // 4, sz[1] // 4
        self.l1 = nn.Sequential(
            *[ResBlk(hid, hid * 2, hid, 5, sz) for _ in range(n_layer)])
        self.l2 = nn.Sequential(
            *[ResBlk(hid, hid * 2, hid, 5, sz2) for _ in range(n_layer)])
        self.l3 = nn.Sequential(
            *[ResBlk(hid, hid * 2, hid, 5, sz4) for _ in range(n_layer)])
        self.l4 = nn.Sequential(
            *[ResBlk(hid, hid * 2, hid, 5, sz4) for _ in range(n_layer)])
        self.l4 = nn.Sequential(
            *[ResBlk(hid, hid * 2, hid, 5, sz4) for _ in range(n_layer)])
        self.l5 = nn.Sequential(*[
            ResBlk(hid * 2, hid * 4, hid * 2, 5, sz2) for _ in range(n_layer)
        ])
        self.l6 = nn.Sequential(*[
            ResBlk(hid * 3, hid * 6, hid * 3, 5, sz) for _ in range(n_layer)
        ])

        self.lout = PixelPredictor(hid * 3, out_ch)

    def _body(self, x: torch.Tensor) -> torch.Tensor:
        x = self.lin(x)

        x1 = self.l1(x)
        x2 = self.l2(x1[..., ::2, ::2])
        x3 = self.l3(x2[..., ::2, ::2])

        x4 = self.l4(x3)
        x5 = self.l5(torch.cat([F.interpolate(x4, scale_factor=2), x2], dim=1))
        x6 = self.l6(torch.cat([F.interpolate(x5, scale_factor=2), x1], dim=1))
        return F.relu(x6)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x6 = self._body(x)
        return self.lout(x6, x)

    def sample_xy(self, x: torch.Tensor, coord_x: int, coord_y: int,
                  temp: float) -> torch.Tensor:
        x6 = self._body(x)
        return self.lout.sample(x6[:, :, coord_y, coord_x], temp)


class PixelCNN(PixCNNBase):
    """
    A PixelCNN model with 6 blocks

    Args:
        hid (int): the number of hidden channels in the blocks
        sz ((int, int)): the size of the images to learn. Must be square
        channels (int): number of channels in the data. 3 for RGB images
    """
    @experimental
    def __init__(self,
                 hid: int,
                 sz: Tuple[int, int],
                 channels: int = 3,
                 n_layer: int = 3) -> None:
        super(PixelCNN, self).__init__(channels,
                                       hid,
                                       channels,
                                       256,
                                       sz,
                                       n_layer=n_layer)
        self.channels = channels

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """A forward pass for training"""
        return super().forward(x)

    def sample(self, temp: float, N: int) -> torch.Tensor:
        """
        Sample a batch of images

        Args:
            temp (float): the sampling temperature
            N (int): number of images to generate in the batch

        Returns:
            A batch of images
        """
        img = torch.zeros(N,
                          self.channels,
                          *self.sz,
                          device=next(self.parameters()).device).uniform_(
                              0, 1)
        return self.sample_(img, temp)

    def partial_sample(self, x: torch.Tensor, temp: float) -> torch.Tensor:
        x[:, :, x.shape[2] // 2, :] = 0
        return self.sample_(x, temp, start_coord=(x.shape[2] // 2, 0))

    def sample_cond(self, cond: torch.Tensor, temp: float) -> torch.Tensor:
        device = next(self.parameters()).device
        img = torch.empty(cond.shape[0],
                          self.sz[0],
                          *self.sz,
                          device=device).uniform_(0, 1)
        cond_rsz = F.interpolate(cond, size=img.shape[2:], mode='nearest')
        img = torch.cat([img, cond_rsz], dim=1)
        return self.sample_(img, temp)[:, cond.shape[1]:]

    def sample_(self,
                img: torch.Tensor,
                temp: float = 0,
                start_coord: Tuple[int, int] = (0, 0)) -> torch.Tensor:
        self.eval()
        with torch.no_grad():
            for row in range(start_coord[0], self.sz[0]):
                for c in range(start_coord[1] if row == start_coord[0] else 0,
                               self.sz[0]):
                    x = self.sample_xy(img * 2 - 1, c, row, temp)
                    img[:, :, row, c] = x

        return img



import torch
import torchelie as tch
import torchelie.utils as tu
from torchelie.recipes.gan import GANRecipe
import torchvision.transforms as TF
import torchelie.loss.gan.standard as gan_loss
from torchelie.loss.gan.penalty import zero_gp
from torchelie.datasets.pix2pix import UnlabeledImages
from torchelie.models import *
import torch.nn as nn


@tu.experimental
def gradient_penalty_M(model, fake, ins, real, objective_norm: float):
    fake = fake.detach()
    ins = ins.detach()

    t = torch.rand(fake.shape[0], 1, 1, 1, device=fake.device)
    fake = t * fake + (1 - t) * real

    fake.requires_grad_(True)
    ins.requires_grad_(True)

    out = F.softmax(model(fake, ins)['matches'], dim=1).sum()

    g = torch.autograd.grad(outputs=out,
                            inputs=fake,
                            create_graph=True,
                            only_inputs=True)[0]

    g_norm = g.pow(2).sum(dim=(1, 2, 3)).add_(1e-8).sqrt()
    return (g_norm - objective_norm).pow(2).mean(), g_norm.mean().item()


class Matcher(nn.Module):
    @tu.experimental
    def __init__(self):
        super().__init__()
        self.net = res_discr_3l()
        self.net.classifier = nn.Sequential()
        self.net.to_equal_lr()
        self.proj_size = 256
        self.proj_A = tu.kaiming(tnn.Conv1x1(128, self.proj_size), dynamic=True)
        self.proj_B = tu.kaiming(tnn.Conv1x1(128, self.proj_size), dynamic=True)

    def forward(self, fake, ins):
        # fake = F.interpolate(fake, scale_factor=0.5, mode='bilinear')
        # ins = F.interpolate(ins, scale_factor=0.5, mode='bilinear')
        f1 = self.proj_A(self.net(fake))
        f2 = self.proj_B(self.net(ins))
        n, c, h, w = f1.shape
        out = torch.bmm(
            f1.permute(2, 3, 0, 1).reshape(-1, n, c),
            f2.permute(2, 3, 0, 1).reshape(-1, n, c).permute(0, 2, 1)
        )
        out = out.view(h, w, n, n).permute(2, 3, 0, 1)

        N = len(out)
        labels = torch.arange(N, device=out.device)
        labels = labels.view(N, 1, 1).expand(N, h, w)
        return {'matches': out,
                'loss': F.cross_entropy(out, labels),
                'labels': labels}


def Crop(x):
    return x.crop((20, 0, 220, 220))


@tu.experimental
def celeba(male, tfm=None):
    from torchvision.datasets import CelebA
    celeba = CelebA('~/.torch/celeba', download=True, target_type=[])
    male_idx = celeba.attr_names.index('Male')
    files = [
        f'~/.torch/celeba/celeba/img_align_celeba/{celeba.filename[i]}'
        for i in range(len(celeba))
        if celeba.attr[i, male_idx] == (1 if male else 0)
    ]
    return tch.datasets.pix2pix.ImagesPaths(files, tfm)


@tu.experimental
def train(rank, world_size):
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('--data-A', required=True)
    parser.add_argument('--data-B', required=True)
    parser.add_argument('--r0-D', default=0.0001, type=float)
    parser.add_argument('--r0-M', default=0.0001, type=float)
    parser.add_argument('--consistency', default=0.01, type=float)
    parser.add_argument('--from-ckpt')
    opts = parser.parse_args()

    G = pix2pix_256()
    D = res_discr_4l()
    D.add_minibatch_stddev()
    D.to_equal_lr()
    M = Matcher()

    for m in G.modules():
        if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
            # tu.kaiming(m, a=0.2, dynamic=True)
            weight_norm_and_equal_lr(m, 0.2)
    G.remove_batchnorm()

    if rank == 0:
        print(G)
        print(D)
        print(M)

    G = torch.nn.parallel.DistributedDataParallel(G.to(rank), [rank], rank)
    D = torch.nn.parallel.DistributedDataParallel(D.to(rank), [rank], rank)
    M = torch.nn.parallel.DistributedDataParallel(M.to(rank), [rank], rank)

    SIZE = 128
    if False:
        ds_A = UnlabeledImages(
            opts.data_A,
            TF.Compose([
                TF.Resize(SIZE),
                TF.CenterCrop(SIZE),
                TF.RandomHorizontalFlip(),
                TF.ToTensor(),
            ]))
        ds_B = UnlabeledImages(
            opts.data_B,
            TF.Compose([
                Crop,
                TF.Resize(SIZE),
                TF.CenterCrop(SIZE),
                TF.RandomHorizontalFlip(),
                TF.ToTensor(),
            ]))
    else:
        ds_A = celeba(
            True,
            TF.Compose([
                TF.Resize(SIZE),
                TF.CenterCrop(SIZE),
                TF.RandomHorizontalFlip(),
                TF.ToTensor(),
            ]))
        ds_B = celeba(
            False,
            TF.Compose([
                TF.Resize(SIZE),
                TF.CenterCrop(SIZE),
                TF.RandomHorizontalFlip(),
                TF.ToTensor(),
            ]))
    print('ds', len(ds_A), len(ds_B))
    ds = tch.datasets.RandomPairsDataset(ds_A, ds_B)

    ds = torch.utils.data.DataLoader(ds,
                                     8,
                                     num_workers=4,
                                     drop_last=True,
                                     shuffle=True,
                                     pin_memory=True)

    def G_fun(batch) -> dict:
        x, y = batch
        D.eval()
        out = G(x * 2 - 1)
        with D.no_sync():
            loss = gan_loss.generated(D(out * 2 - 1))
        loss.backward(retain_graph=True)

        with M.no_sync():
            clf_loss = opts.consistency * M(out * 2 - 1, x * 2 - 1)['loss']

        # labels = torch.arange(len(matches), device=matches.device)
        # clf_loss = opts.consistency * F.cross_entropy( matches, torch.arange(len(matches), device=matches.device))
        clf_loss.backward()
        return {'G_loss': loss.item()}

    class GradientPenalty:
        def __init__(self, gamma):
            self.gamma = gamma
            self.iters = 0
            self.last_norm = float('nan')

        def __call__(self, model, real, fake):
            if self.iters < 100 or self.iters % 4 == 0:
                real = real.detach()
                fake = fake.detach()
                gp, g_norm = zero_gp(model, real, fake)
                # Sync the gradient on the next backward
                if torch.any(torch.isnan(gp)):
                    gp.detach_()
                else:
                    (4 * self.gamma * gp).backward()
                self.last_norm = g_norm
            self.iters += 1
            return self.last_norm

    gradient_penalty = GradientPenalty(opts.r0_D)

    def D_fun(batch) -> dict:
        G.eval()
        x, y = batch
        with G.no_sync():
            with torch.no_grad():
                out = G(x * 2 - 1)
        fake = out * 2 - 1
        real = y * 2 - 1
        with D.no_sync():
            prob_fake = D(fake)
            fake_correct = prob_fake.detach().lt(0).int().eq(1).sum()
            fake_loss = gan_loss.fake(prob_fake)
            fake_loss.backward()

        with D.no_sync():
            g_norm = gradient_penalty(D, real, fake)

        prob_real = D(real)
        real_correct = prob_real.detach().gt(0).int().eq(1).sum()
        real_loss = gan_loss.real(prob_real)
        real_loss.backward()

        with M.no_sync():
            gp, match_g_norm = gradient_penalty_M(M, fake, x * 2 - 1, real, 0)
            (opts.r0_M * gp).backward()

        M_out = M(fake, x * 2 - 1)
        matches = M_out['matches']
        labels = M_out['labels']
        match_correct = matches.argmax(1).eq(labels).float().mean()

        loss = M_out['loss']
        loss.backward()

        return {
            'out': out,
            'fake_loss': fake_loss.item(),
            'prob_fake': torch.sigmoid(prob_fake).mean().item(),
            'prob_real': torch.sigmoid(prob_real).mean().item(),
            'real_loss': real_loss.item(),
            'g_norm': g_norm,
            'D-correct':
            (fake_correct + real_correct) / (2 * prob_fake.numel()),
            'match_correct': match_correct,
            'match_g_norm': match_g_norm,
        }

    def test_fun(_):
        return {}

    recipe = GANRecipe(G,
                       D,
                       G_fun,
                       D_fun,
                       test_fun,
                       ds,
                       checkpoint='face_inpaint' if rank == 0 else None,
                       visdom_env='main' if rank == 0 else None)
    recipe.register('M', M)

    recipe.callbacks.add_callbacks([
        tch.callbacks.Optimizer(
            tch.optim.RAdamW(D.parameters(),
                             lr=2e-3,
                             betas=(0., 0.99),
                             weight_decay=0)),
        tch.callbacks.Optimizer(
            tch.optim.RAdamW(M.parameters(),
                             lr=2e-3,
                             betas=(0., 0.99),
                             weight_decay=0)),
        tch.callbacks.Log('out', 'out'),
        tch.callbacks.Log('batch.0', 'x'),
        tch.callbacks.Log('batch.1', 'y'),
        # tch.callbacks.Log('batch.0.1', 'y'),
        tch.callbacks.WindowedMetricAvg('fake_loss', 'fake_loss'),
        tch.callbacks.WindowedMetricAvg('real_loss', 'real_loss'),
        tch.callbacks.WindowedMetricAvg('prob_fake', 'prob_fake'),
        tch.callbacks.WindowedMetricAvg('prob_real', 'prob_real'),
        tch.callbacks.WindowedMetricAvg('D-correct', 'D-correct'),
        tch.callbacks.WindowedMetricAvg('match_correct', 'match_correct'),
        tch.callbacks.Log('g_norm', 'g_norm'),
        tch.callbacks.Log('match_g_norm', 'match_g_norm'),
    ])
    recipe.G_loop.callbacks.add_callbacks([
        tch.callbacks.Optimizer(
            tch.optim.RAdamW(G.parameters(),
                             lr=2e-3,
                             betas=(0., 0.99),
                             weight_decay=0)),
    ])
    recipe.to(rank)
    if opts.from_ckpt is not None:
        recipe.load_state_dict(torch.load(opts.from_ckpt, map_location='cpu'))
    recipe.run(200)


if __name__ == '__main__':
    tu.parallel_run(train)



import torch
import torch.nn as nn
from typing import Optional


class Noise(nn.Module):
    """
    Add gaussian noise to the input, with a per channel or global learnable std.

    Args:
        ch (int): number of input channels for a different std on each channel,
            or 1
    """
    def __init__(self, ch: int, inplace: bool = False, bias: bool = False):
        super(Noise, self).__init__()
        self.a = nn.Parameter(torch.zeros(ch, 1, 1))
        self.inplace = inplace
        self.bias = nn.Parameter(torch.zeros_like(self.a)) if bias else None

    def forward(self,
                x: torch.Tensor,
                z: Optional[torch.Tensor] = None) -> torch.Tensor:
        N, C, H, W = x.shape
        if z is None:
            z = torch.randn(N, 1, H, W, device=x.device, dtype=x.dtype)
        else:
            assert z.shape == [N, 1, H, W]
        z = z * self.a
        if self.bias is not None:
            z.add_(self.bias)

        if self.inplace:
            return x.add_(z)
        else:
            return x + z



from setuptools import setup, find_packages

setup(
        name='Torchelie',
        version='0.1dev',
        packages=find_packages(),
        classifiers=[
            "License :: OSI Approved :: MIT License",
        ],
        install_requires=[
            'visdom>=0.1.8',
            'crayons>=0.2',
            'torchvision>=0.4',
            'numpy>=1.16',
            'Pillow>=6',
        ]
)



from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
from scvi import REGISTRY_KEYS, settings
from scvi.module.base import BaseModuleClass, LossOutput, auto_move_data
from torch import nn

from ._constants import LOSS_KEYS
from ._utils import _logger, tensor2mat
from .Torchelie.torchelie.loss import PerceptualLoss


class DoubleConv(nn.Module):
    def __init__(
        self, in_ch: int, out_ch: int, kernel_size: int = 3, padding_size: int = 1
    ):
        super(DoubleConv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, kernel_size, padding=padding_size),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, kernel_size, padding=padding_size),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
        )

    def forward(self, input):
        return self.conv(input)


class DownConv(nn.Module):
    def __init__(
        self, in_ch: int, out_ch: int, kernel_size: int = 3, padding_size: int = 1
    ):
        super(DownConv, self).__init__()
        self.conv = nn.Sequential(
            nn.MaxPool2d(2), DoubleConv(in_ch, out_ch, kernel_size, padding_size)
        )

    def forward(self, input):
        return self.conv(input)


class UpConv(nn.Module):
    def __init__(
        self, in_ch: int, out_ch: int, kernel_size: int = 3, padding_size: int = 1
    ):
        super(UpConv, self).__init__()
        self.up = nn.ConvTranspose2d(in_ch, out_ch, 2, stride=2)
        self.conv = DoubleConv(in_ch, out_ch, kernel_size, padding_size)

    def forward(self, input1, input2):
        return self.conv(torch.cat([self.up(input1), input2], dim=1))


class HiClipModule(BaseModuleClass):
    def __init__(
        self,
        main_loc: str,
        sub_loc: str,
        target_loc: str,
        main_channel: List = [1, 64, 128, 256, 128, 64],
        sub_channel: List = [1, 16, 32, 64],
        out_channel: List = [128, 64, 1],
        main_conv: List = [
            (3, 1),
            ((5, 1), (2, 0)),
            ((1, 5), (0, 2)),
            ((5, 1), (2, 0)),
            ((1, 5), (0, 2)),
        ],
        sub_conv: List = [(1, 0), (1, 0), (1, 0)],
        out_conv: List = [(3, 1), (1, 0)],
        seed: int = 0,
    ):
        super().__init__()

        assert len(main_channel) == len(main_conv) + 1
        assert len(sub_channel) == len(sub_conv) + 1
        assert len(out_channel) == len(out_conv) + 1

        assert len(main_channel) % 2 == 0
        assert main_channel[-1] + sub_channel[-1] == out_channel[0]

        torch.manual_seed(seed)
        np.random.seed(seed)
        settings.seed = seed

        self.main_loc, self.sub_loc, self.target_loc = main_loc, sub_loc, target_loc

        self.main_depths = int((len(main_conv) - 1) / 2)
        self.main_modules = [
            DoubleConv(
                main_channel[0], main_channel[1], main_conv[0][0], main_conv[0][1]
            )
        ]
        for i in range(self.main_depths):
            _i = 1 + i
            self.main_modules.append(
                DownConv(
                    main_channel[_i],
                    main_channel[_i + 1],
                    main_conv[_i][0],
                    main_conv[_i][1],
                )
            )
        for i in range(self.main_depths):
            _i = 1 + self.main_depths + i
            self.main_modules.append(
                UpConv(
                    main_channel[_i],
                    main_channel[_i + 1],
                    main_conv[_i][0],
                    main_conv[_i][1],
                )
            )
        self.main_modules = nn.ModuleList(self.main_modules)
        self.sub_modules = nn.ModuleList(
            [
                DoubleConv(
                    sub_channel[i], sub_channel[i + 1], sub_conv[i][0], sub_conv[i][1]
                )
                for i in range(len(sub_conv))
            ]
        )
        self.out_modules = nn.ModuleList(
            [
                DoubleConv(
                    out_channel[i], out_channel[i + 1], out_conv[i][0], out_conv[i][1]
                )
                for i in range(len(out_conv))
            ]
        )

        self.L1_loss = nn.L1Loss(reduction="mean")
        self.perceptual_loss_layers = [
            "conv1_1",
            "relu1_1",
            "conv1_2",
            "relu1_2",
            "maxpool1",
            "conv2_1",
            "relu2_1",
            "conv2_2",
            "relu2_2",
            "maxpool2",
            "conv3_1",
            "relu3_1",
            "conv3_2",
            "relu3_2",
            "conv3_3",
            "relu3_3",
            "conv3_4",
            "relu3_4",
            "maxpool3",  # noqa: E131
            "conv4_1",
            "relu4_1",
            "conv4_2",
            "relu4_2",
            "conv4_3",
            "relu4_3",
            "conv4_4",
            "relu4_4",
            "maxpool4",  # noqa: E131
            "conv5_1",
            "relu5_1",
            "conv5_2",
            "relu5_2",
            "conv5_3",
            "relu5_3",
            "conv5_4",
            "relu5_4",  # 'maxpool5'
        ]

    def _get_inference_input(self, tensors: Dict[Any, Any], **kwargs):
        main = tensors[self.main_loc]  # [batch_size, interactions]
        sub = tensors[self.sub_loc]  # [batch_size, interactions]
        # sample_indices = tensors[REGISTRY_KEYS.INDICES_KEY].long().ravel()

        main = torch.unsqueeze(tensor2mat(main), 1)
        sub = torch.unsqueeze(tensor2mat(sub), 1)

        input_dict = {
            "main": main,
            "sub": sub,
        }
        return input_dict

    @auto_move_data
    def inference(
        self,
        main: torch.Tensor,
        sub: torch.Tensor,
    ):
        # main = self.main_modules[0](main)
        mains = []
        for i in range(self.main_depths):
            main = self.main_modules[i](main)
            mains.append(main)
        main = self.main_modules[self.main_depths](main)
        for i in range(self.main_depths):
            main = self.main_modules[1 + self.main_depths + i](main, mains[-i - 1])

        for module in self.sub_modules:
            sub = module(sub)

        out = torch.cat([main, sub], dim=1)
        for module in self.out_modules:
            out = module(out)

        out = nn.ReLU()(out)
        return out

    @auto_move_data
    def loss(
        self,
        tensors: Dict[str, torch.Tensor],
        inference_output: torch.Tensor,
    ) -> Dict[str, float]:
        losses = dict()
        target = tensors[self.target_loc]
        target = torch.unsqueeze(tensor2mat(target), 1)

        target_is_0, target_is_not_0 = target == 0, target != 0

        losses[LOSS_KEYS.L1] = self.L1_loss(
            inference_output[target_is_not_0], target[target_is_not_0]
        )

        _output, _target = inference_output.clone().detach(), target.clone().detach()
        _output[target_is_0], _target[target_is_0] = 0, 0
        losses[LOSS_KEYS.PERCEPTUAL] = PerceptualLoss(
            self.perceptual_loss_layers,
            rescale=True,
        ).to(inference_output.device)(_output, _target)

        return losses

    @auto_move_data
    def forward(
        self,
        tensors,
        compute_loss=True,
    ) -> Union[tuple[torch.Tensor], tuple[torch.Tensor, LossOutput]]:
        inference_inputs = self._get_inference_input(tensors)
        inference_output = self.inference(**inference_inputs)

        if compute_loss:
            losses = self.loss(tensors, inference_output)
            return inference_output, losses
        else:
            return inference_output



import math
import torch.nn as nn
import torchelie.nn as tnn
import torchelie.utils as tu


@tu.experimental
class MBConv(nn.Module):

    def __init__(self,
                 in_ch: int,
                 out_ch: int,
                 ks: int,
                 stride: int = 1,
                 mul_factor: int = 6):
        super(MBConv, self).__init__()
        self.in_ch = in_ch
        self.out_ch = out_ch
        self.ks = ks
        self.stride = stride
        self.factor = mul_factor

        hid = in_ch * mul_factor
        self.branch = tnn.CondSeq(
            tu.xavier(tnn.Conv1x1(in_ch, hid, bias=False)), nn.BatchNorm2d(hid),
            tnn.HardSwish(),
            tu.xavier(
                nn.Conv2d(hid,
                          hid,
                          ks,
                          stride=stride,
                          padding=ks // 2,
                          groups=hid,
                          bias=False)), nn.BatchNorm2d(hid), tnn.HardSwish(),
            tnn.SEBlock(hid, reduction=4), tu.xavier(tnn.Conv1x1(hid, out_ch)),
            nn.BatchNorm2d(out_ch))

        self.shortcut = tnn.CondSeq()

        if stride != 1:
            self.shortcut.add_module(
                'pool', nn.AvgPool2d(stride, stride, ceil_mode=True))

        if in_ch != out_ch:
            self.shortcut.add_module('conv',
                                     tnn.Conv1x1(in_ch, out_ch, bias=False))
            self.shortcut.add_module('bn', nn.BatchNorm2d(out_ch))

    def __repr__(self):
        return "MBConv({}, {}, factor={}, k{}x{}s{}))".format(
            self.in_ch, self.out_ch, self.factor, self.ks, self.ks, self.stride)

    def forward(self, x):
        return self.branch(x).add_(self.shortcut(x))


class EfficientNet(tnn.CondSeq):

    @tu.experimental
    def __init__(self, in_ch, num_classes, B=0):

        def ch(ch):
            return int(ch * 1.1**B) // 8 * 8

        def n_layers(d):
            return int(math.ceil(d * 1.2**B))

        def r():
            return int(224 * 1.15**B)

        super(EfficientNet, self).__init__(
            # Stage 1
            # nn.UpsamplingBilinear2d(size=(r(), r())),
            tu.kaiming(tnn.Conv3x3(in_ch, ch(32), stride=2, bias=False)),
            nn.BatchNorm2d(ch(32)),
            tnn.HardSwish(),

            # Stage 2
            MBConv(ch(32), ch(16), 3, mul_factor=1),
            *[
                MBConv(ch(16), ch(16), 3, mul_factor=1)
                for _ in range(n_layers(1) - 1)
            ],

            # Stage 3
            MBConv(ch(16), ch(24), 3, stride=2),
            *[MBConv(ch(24), ch(24), 3) for _ in range(n_layers(2) - 1)],

            # Stage 4
            MBConv(ch(24), ch(40), 5, stride=2),
            *[MBConv(ch(40), ch(40), 5) for _ in range(n_layers(2) - 1)],

            # Stage 5
            MBConv(ch(40), ch(80), 3, stride=2),
            *[MBConv(ch(80), ch(80), 3) for _ in range(n_layers(3) - 1)],

            # Stage 6
            MBConv(ch(80), ch(112), 5),
            *[MBConv(ch(112), ch(112), 5) for _ in range(n_layers(3) - 1)],

            # Stage 7
            MBConv(ch(112), ch(192), 5, stride=2),
            *[MBConv(ch(192), ch(192), 5) for _ in range(n_layers(4) - 1)],

            # Stage 8
            MBConv(ch(192), ch(320), 3),
            *[MBConv(ch(320), ch(320), 3) for _ in range(n_layers(1) - 1)],
            tu.kaiming(tnn.Conv1x1(ch(320), ch(1280), bias=False)),
            nn.BatchNorm2d(ch(1280)),
            tnn.HardSwish(),
            nn.AdaptiveAvgPool2d(1),
            tnn.Reshape(-1),
            tu.xavier(nn.Linear(ch(1280), num_classes)))



from .trainandtest import TrainAndTest


def TrainAndCall(model,
                 train_fun,
                 test_fun,
                 train_loader,
                 test_every=100,
                 visdom_env='main',
                 checkpoint='model',
                 log_every=10,
                 key_best=None):
    """
    Train a model and evaluate it with a custom function. The model is
    automatically registered and checkpointed as :code:`checkpoint['model']`,
    and put in eval mode when testing.

    Training callbacks:

    - Counter for counting iterations, connected to the testing loop as well
    - VisdomLogger
    - StdoutLogger

    Testing:

    Testing loop is in :code:`.test_loop`.

    Testing callbacks:

    - VisdomLogger
    - StdoutLogger
    - Checkpoint

    Args:
        model (nn.Model): a model
        train_fun (Callabble): a function that takes a batch as a single
            argument, performs a training step and return a dict of values to
            populate the recipe's state.
        test_fun (Callable): a function taking no argument that performs
            something to evaluate your model and returns a dict to populate the
            state.
        train_loader (DataLoader): Training set dataloader
        test_every (int): testing frequency, in number of iterations (default:
            100)
        visdom_env (str): name of the visdom environment to use, or None for
            not using Visdom (default: None)
        checkpoint (str): checkpointing path or None for no checkpointing
        log_every (int): logging frequency, in number of iterations (default:
            10)
        key_best (function or None): a key function for comparing states.
            Checkpointing the greatest.

    Returns:
        a configured Recipe
    """

    def test_fun_wrap(_):
        return test_fun()

    return TrainAndTest(model,
                        train_fun,
                        test_fun_wrap,
                        train_loader=train_loader,
                        test_loader=range(1),
                        test_every=test_every,
                        visdom_env=visdom_env,
                        checkpoint=checkpoint,
                        log_every=log_every,
                        key_best=key_best)



from torch.utils.data import Dataset
from torchelie.utils import indent
from typing import List, Tuple, Any, Generic, TypeVar, Sequence
from typing import overload

T = TypeVar('T')
U = TypeVar('U')


class CatedSamples(Generic[T]):
    def __init__(self, samples: List[List[Tuple[T, int]]]) -> None:
        self.samples = samples
        self.n_classes = [
            len(set(samp[1] for samp in sample)) for sample in samples
        ]

    def __len__(self) -> int:
        return sum(len(ds) for ds in self.samples)

    def __getitem__(self, i: int) -> Tuple[T, int]:
        class_offset = 0
        for samp, n_class in zip(self.samples, self.n_classes):
            if i < len(samp):
                return samp[i][0], samp[i][1] + class_offset
            i -= len(samp)
            class_offset += n_class
        raise IndexError


class CatedLists(Sequence[T]):
    def __init__(self, ls: List[List[T]]) -> None:
        self.ls = ls

    def __len__(self) -> int:
        return sum([len(ds) for ds in self.ls])

    @overload
    def __getitem__(self, i: int) -> T:
        ...

    @overload
    def __getitem__(self, i: slice) -> Sequence[T]:
        ...

    def __getitem__(self, i):
        if isinstance(i, slice):
            return [self[ii] for ii in range(*i.indices(len(self)))]

        for catedlist in self.ls:
            if i < len(catedlist):
                return catedlist[i]
            i -= len(catedlist)
        raise IndexError


class HorizontalConcatDataset(Dataset):
    """
    Concatenates multiple datasets. However, while torchvision's ConcatDataset
    just concatenates samples, torchelie's also relabels classes. While a
    vertical concat like torchvision's is useful to add more examples per
    class, an horizontal concat merges datasets to more classes.

    Args:
        datasets (list of Dataset): the datasets to concatenate
    """
    classes: CatedLists[str]

    def __init__(self, datasets: List) -> None:
        self.datasets = datasets

        self.classes = CatedLists([ds.classes for ds in datasets])
        self.samples = CatedSamples([ds.samples for ds in datasets])
        self.class_to_idx = {nm: i for i, nm in enumerate(self.classes)}

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, i: int) -> Tuple[Any, int]:
        class_offset = 0
        for ds in self.datasets:
            if i < len(ds):
                x, t = ds[i]
                return x, t + class_offset
            i -= len(ds)
            class_offset += len(ds.classes)
        raise IndexError

    def __repr__(self) -> str:
        return "DatasetConcat:\n" + '\n--\n'.join(
            [indent(repr(d)) for d in self.datasets])


class MergedSamples:
    def __init__(self, ds) -> None:
        self.ds = ds

    def __len__(self) -> int:
        return sum(len(d) for d in self.ds.datasets)

    def __getitem__(self, i: int):
        for ds in self.ds.datasets:
            if i < len(ds):
                x, y, *ys = ds.samples[i]
                return [x, self.ds.class_to_idx[ds.classes[y]]] + ys
            i -= len(ds)
        raise IndexError


class MergedDataset(Dataset):
    def __init__(self, datasets):
        self.datasets = datasets
        self.classes = list(set(c for d in datasets for c in d.classes))
        self.classes.sort()
        self.class_to_idx = {c: i for i, c in enumerate(self.classes)}

        self.samples = MergedSamples(self)

    def __len__(self):
        return sum(len(d) for d in self.datasets)

    def __getitem__(self, i):
        for ds in self.datasets:
            if i < len(ds):
                x, y, *ys = ds[i]
                return [x, self.class_to_idx[ds.classes[y]]] + ys
            i -= len(ds)
        raise IndexError

    def __repr__(self):
        return "MergedDatasets: \n" + "\n".join(
            [indent(repr(ds)) for ds in self.datasets])



import sys
import argparse

import crayons

import torch
import torch.nn as nn
import torch.nn.functional as F

from torchvision.datasets import MNIST, CIFAR10
import torchvision.transforms as TF

import torchelie.nn as tnn
import torchelie.models
import torchelie as tch
from torchelie.models import ClassCondResNetDebug
from torchelie.utils import nb_parameters
from torchelie.recipes.classification import Classification
from torchelie.optim import RAdamW

parser = argparse.ArgumentParser()
parser.add_argument('--cpu', action='store_true')
parser.add_argument('--dataset',
                    type=str,
                    choices=['mnist', 'cifar10'],
                    default='mnist')
parser.add_argument('--models', default='all')
opts = parser.parse_args()

device = 'cpu' if opts.cpu else 'cuda'


class TrueOrFakeLabelDataset:
    def __init__(self, dataset):
        self.dataset = dataset
        self.classes = ['Fake', 'True']

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, i):
        x, y = self.dataset[i]
        if torch.randn(1).item() < 0:
            return x, 1, y
        return x, 0, torch.randint(0, 10, (1, )).item()


tfms = TF.Compose([TF.Resize(32), TF.ToTensor()])
ds = TrueOrFakeLabelDataset(
    MNIST('~/.cache/torch/mnist', download=True, transform=tfms))
dt = TrueOrFakeLabelDataset(
    MNIST('~/.cache/torch/mnist', download=True, transform=tfms, train=False))
dl = torch.utils.data.DataLoader(ds,
                                 num_workers=4,
                                 batch_size=32,
                                 shuffle=True)
dlt = torch.utils.data.DataLoader(dt,
                                  num_workers=4,
                                  batch_size=32,
                                  shuffle=True)


def train_net():
    model = ClassCondResNetDebug(2, 10, in_ch=1)

    def train_step(batch):
        x, y, z = batch

        out = model(x, z)
        loss = F.cross_entropy(out, y)
        loss.backward()

        return {'loss': loss, 'pred': out}

    def validation_step(batch):
        x, y, z = batch

        out = model(x, z)
        loss = F.cross_entropy(out, y)
        return {'loss': loss, 'pred': out}

    clf = Classification(model, train_step, validation_step, dl, dlt,
                         ds.classes).to(device)
    clf.callbacks.add_callbacks(
        [tch.callbacks.Optimizer(tch.optim.RAdamW(model.parameters()))])
    clf.run(2)


train_net()



import torch
import torch.nn as nn
import torch.nn.functional as F

from torchelie.utils import experimental
import torchelie.nn as tnn
from torchelie.utils import xavier
from typing import List


@experimental
class AutoGAN(nn.Module):
    """
    Generator discovered in AutoGAN: Neural Architecture Search for Generative
    Adversarial Networks.

    Args:
        arch (list): architecture specification: a list of output channel for
            each block. Each block doubles the resolution of the generated
            image. Example: `[512, 256, 128, 64, 32]`.
        n_skip_max (int): how many blocks far back will be used for the skip
            connections maximum.
        in_noise (int): dimension of the input noise vector
        out_ch (int): number of channels on the image
        batchnorm_in_output (bool): whether to have a batchnorm just before
            projecting to RGB. I have found it better on False, but the
            official AutoGAN repo has it.
    """
    n_skip_max: int
    make_noise: nn.Module
    blocks: nn.ModuleList
    to_rgb: nn.Sequential

    def __init__(self,
                 arch: List[int],
                 n_skip_max: int = 2,
                 in_noise: int = 256,
                 out_ch: int = 3,
                 batchnorm_in_output: bool = False) -> None:
        super().__init__()
        self.n_skip_max = n_skip_max
        self.make_noise = xavier(nn.Linear(in_noise, 4 * 4 * arch[0]))

        in_ch = arch[0]
        blocks = []
        lasts: List[int] = []
        for i, out in enumerate(arch[1:]):
            mode = 'nearest' if i % 2 == 0 else 'bilinear'
            blocks.append(tnn.AutoGANGenBlock(in_ch, out, lasts, mode=mode))
            lasts = ([out] + lasts)[:n_skip_max]
            in_ch = out
        self.blocks = nn.ModuleList(blocks)
        if batchnorm_in_output:
            self.to_rgb = nn.Sequential(nn.BatchNorm2d(arch[-1]),
                                        nn.ReLU(True),
                                        xavier(tnn.Conv3x3(arch[-1], out_ch)))
        else:
            self.to_rgb = nn.Sequential(nn.ReLU(True),
                                        xavier(tnn.Conv3x3(arch[-1], out_ch)))

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        """
        Forward pass

        Args:
            z (tensor): A batch of noise vectors

        Returns:
            generated batch of images
        """
        x = self.make_noise(z)
        x = x.view(x.shape[0], -1, 4, 4)

        skips: List[torch.Tensor] = []
        for b in self.blocks:
            x, sk = b(x, skips)
            skips = ([sk] + skips)[:self.n_skip_max]
        return torch.sigmoid(self.to_rgb(F.leaky_relu(x, 0.2)))


@experimental
def autogan_128(in_noise: int, out_ch: int = 3) -> AutoGAN:
    return AutoGAN(arch=[512, 512, 256, 128, 64, 32],
                   n_skip_max=3,
                   in_noise=in_noise,
                   out_ch=out_ch)


@experimental
def autogan_64(in_noise: int, out_ch: int = 3) -> AutoGAN:
    return AutoGAN(arch=[512, 256, 128, 64, 32],
                   n_skip_max=3,
                   in_noise=in_noise,
                   out_ch=out_ch)


@experimental
def autogan_32(in_noise: int, out_ch: int = 3) -> AutoGAN:
    return AutoGAN(arch=[256, 128, 64, 32],
                   n_skip_max=3,
                   in_noise=in_noise,
                   out_ch=out_ch)



import logging
from math import log10, sqrt

import numpy as np
import torch
from numpy.fft import fft2, ifft2, ifftshift
from scvi import scvi_logger, settings

"""
Hi-C matrix transformation
"""


def tensor2mat(tensor: torch.Tensor) -> torch.Tensor:
    len_sqrt = int(sqrt(tensor.size()[-1]))
    new_x_shape = tensor.size()[:-1] + (len_sqrt, len_sqrt)
    tensor = tensor.view(*new_x_shape)

    return tensor


"""
Coarsen the matrix.
"""


def _fftblur(img, sigma):
    h, w = img.shape

    X, Y = np.meshgrid(np.arange(w), np.arange(h))
    X, Y = X - w // 2, Y - h // 2
    Z = np.exp(-0.5 * (X**2 + Y**2) / (sigma**2))
    Z = Z / Z.sum()

    out = ifftshift(ifft2(fft2(img) * fft2(Z)))
    return out


def _fan(sparse_mat, mask):
    N = np.prod(mask.shape)
    num_kept = np.nonzero(mask)[0].shape[0]
    sigma = sqrt(N / (np.pi * num_kept))

    c = _fftblur(sparse_mat, sigma)
    i = _fftblur(mask, sigma)

    mat = np.abs(c / i)
    return mat


def _sparsify(mat, sparse_mat, span):
    h, w = mat.shape
    for i in range(0, h, span):
        for j in range(0, w, span):
            tmp = mat[i : i + span, j : j + span].mean()
            _x, _y = i + span / 2, j + span / 2
            _length = max(abs(_x - _y), 1)
            tmp *= max(log10(_length), 1)
            sparse_mat[i : i + span, j : j + span] += tmp
    return sparse_mat


def coarsen(mat, spans=[5, 10]):
    mask = np.ones(mat.shape)
    sparse_mat = np.zeros(mat.shape)

    spans.sort(reverse=True)
    for span in spans:
        sparse_mat = _sparsify(mat, sparse_mat, span)

    return _fan(sparse_mat, mask)


"""
metric
"""


def hiclip_metric(
    l1_loss: float,
    perceptual_loss: float,
) -> float:
    return np.nanmean([l1_loss, perceptual_loss])


"""
log
"""

_logger = scvi_logger
settings.logging_dir = "./.hiclip/"



import math
import random
import torch
import torch.nn.functional as F
import numpy as np
from typing import Optional, cast, List

# FIXME: The API is not that great, improve it.


def _rollx(img, begin):
    return torch.cat([img[..., :, begin:], img[..., :, :begin]], dim=-1)


def _rolly(img, begin):
    return torch.cat([img[..., begin:, :], img[..., :begin, :]], dim=-2)


def roll(img, x_roll, y_roll):
    """
    Wrap an image

    Args:
        img (3D or 4D image(s) tensor): an image tensor
        x_roll (int): how many pixels to roll on the x axis
        y_roll (int): how many pixels to roll on the y axis

    Returns:
        The rolled tensor
    """
    return _rollx(_rolly(img, y_roll), x_roll)


def center_crop(batch, size):
    """
    Crop the center of a 4D images tensor

    Args:
        batch (4D images tensor): the tensor to crop
        size ((int, int)): size of the resulting image as (height, width)

    Returns:
        The cropped image
    """
    if batch.shape[-2] == size[0] and batch.shape[-1] == size[1]:
        return batch
    y_off = (batch.shape[-2] - size[0]) // 2
    x_off = (batch.shape[-1] - size[1]) // 2
    return batch[..., y_off:y_off + size[0], x_off:x_off + size[1]]


def crop(img, warped=True, sub_img_factor=2):
    """
    Randomly crop a `sub_img_factor` smaller part of `img`.

    Args:
        img (3D or 4D image(s) tensor): input image(s)
        warped (bool): Whether the image should be considered warped (default:
            True)
        sub_img_factor (float): fraction of the image to take. For instance, 2
            will crop a quarter of the image (half the width, half the height).
            (default: 2)
    """
    sz = img.shape
    h = np.random.randint(sz[-2] // sub_img_factor, sz[-2])
    w = np.random.randint(sz[-1] // sub_img_factor, sz[-1])

    if warped:
        y = np.random.randint(sz[-2])
        x = np.random.randint(sz[-1])
        out = _rolly(_rollx(img, x), y)
        return out[..., :h, :w]
    else:
        y = np.random.randint(sz[-2] - h)
        x = np.random.randint(sz[-1] - w)
        return img[:, y:y + h, x:x + w]


def _gblur_kernel_2d(c):
    gaussian = np.array([[1, 2, 1], [2, 4, 2], [1, 2, 1]]) / 16
    kernel = np.zeros((3, 3, 3))
    kernel[c, :, :] = gaussian
    return kernel


def _gblur_kernel():
    return np.stack(
        [_gblur_kernel_2d(0),
         _gblur_kernel_2d(1),
         _gblur_kernel_2d(2)]).astype(np.float32)


def gblur(input):
    """
    Gaussian blur with kernel size 3

    Args:
        input (3D or 4D image(s) tensor): input image

    Returns:
        the blurred tensor
    """
    print('gblur', input.dtype, torch.from_numpy(_gblur_kernel()).dtype)
    return F.conv2d(input,
                    torch.from_numpy(_gblur_kernel()).to(input.device),
                    padding=1)


def _mblur_kernel_2d(c):
    gaussian = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]) / 8
    kernel = np.zeros((3, 3, 3))
    kernel[c, :, :] = gaussian
    return kernel


def _mblur_kernel():
    return np.stack(
        [_mblur_kernel_2d(0),
         _mblur_kernel_2d(1),
         _mblur_kernel_2d(2)]).astype(np.float32)


class BinomialFilter2d(torch.nn.Module):
    def __init__(self, stride: int):
        super().__init__()
        self.stride = stride
        self.register_buffer(
            'weight',
            torch.tensor([[[1.0, 2, 1], [2, 4, 2], [1, 2, 1]]]) / 16)

    def forward(self, x):
        x = torch.nn.functional.pad(x, (1, 1, 1, 1), mode='replicate')
        return torch.nn.functional.conv2d(x,
                                          self.weight.expand(x.shape[1], 1, -1, -1),
                                          groups=x.shape[1],
                                          stride=self.stride,
                                          padding=0)


def mblur(input):
    """
    Mean (or average) blur with kernel size 3

    Args:
        input (3D or 4D image(s) tensor): input image

    Returns:
        the blurred tensor
    """
    return F.conv2d(input,
                    torch.from_numpy(_mblur_kernel()).to(input.device),
                    padding=1)


class AllAtOnceGeometric:
    """
    Various geometric transforms packed up into an affine transformation
    matrix. Transformations can be stacked then applied on a 4D tensor to
    reduce artifact, memory usage, and compute. Fully differentiable.

    >>> img = torch.randn(10, 3, 32, 32)
    >>> transformed = AllAtOnceGeometric(10)\
            .translate(5, 5).scale(0.9).apply(img)

    Note: the transformations get sampled at creation, so that each call to
    apply() runs the same transforms. Construct another AllAtOnceGeometric
    object for another set of transform. This allows to easily run the same
    transforms on paired datasets.

    Note2: Each transform has a prob argument which specifies whether to use
    the transform or bypass it. This makes it easy to implement StyleGAN2-ADA.

    Args:
        B (int): batch size
        init (torch.Tensor): an initial user supplied transformation matrix. If
            not provided, default to identity.
    """
    B: int
    m: torch.Tensor

    def __init__(self, B: int, init: Optional[torch.Tensor] = None) -> None:
        if init is None:
            self.m = torch.stack([torch.eye(3, 3) for _ in range(B)], dim=0)
        else:
            self.m = init
        self.B = B

    def _mix(self, m: torch.Tensor, prob: float) -> None:
        RND = (m.shape[0], 1, 1)
        self.m = torch.where(
            torch.rand(RND) < prob, torch.bmm(m, self.m), self.m)

    def translate(self, x: float, y: float, prob: float = 1.) -> 'AllAtOnceGeometric':
        """
        Randomly translate image horizontally with an offset sampled in [-x, x]
        and vertically [-y, y]. Note that the coordinate are not pixel
        coordinate but texel coordinate between [-1, 1]
        """
        tfm = torch.tensor([[[1, 0., random.uniform(-x, x)],
                             [0, 1, random.uniform(-y, y)], [0, 0, 1]]
                            for _ in range(self.B)])
        self._mix(tfm, prob)
        return self

    def scale(self, x: float, y: float, prob: float = 1.) -> 'AllAtOnceGeometric':
        """
        Randomly scale the image horizontally by a factor [1 - x; 1 + x] and
        vertically by a factor of [1 - y; 1 + y].

        Args:
            x (float): horizontal factor
            y (float): vertical factor
        """
        tfm = torch.tensor([[[random.uniform(1 - x, 1 + x), 0., 0],
                             [0, random.uniform(1 - y, 1 + y), 0], [0, 0, 1]]
                            for _ in range(self.B)])
        self._mix(tfm, prob)
        return self

    def rotate(self, theta: float, prob: float = 1.) -> 'AllAtOnceGeometric':
        """
        Rotate the image by an angle randomly sampled between [-theta, theta]

        Args:
            theta (float): an angle in degrees
        """
        theta = theta * 3.14 / 180
        rot = []
        for _ in range(self.B):
            t = random.uniform(-theta, theta)
            rot.append([[math.cos(t), math.sin(-t), 0],
                        [math.sin(t), math.cos(t), 0], [0, 0, 1]])
        tfm = torch.tensor(rot)
        self._mix(tfm, prob)
        return self

    def flip_x(self, p: float, prob: float = 1.) -> 'AllAtOnceGeometric':
        tfm = []
        for _ in range(self.B):
            p_rot = math.copysign(1, random.random() - (1 - p))
            tfm.append([[p_rot, 0., 0], [0, 1, 0], [0, 0, 1]])
        tfm_matrix = torch.tensor(tfm)
        self._mix(tfm_matrix, prob)
        return self

    def flip_y(self, p: float, prob: float = 1.) -> 'AllAtOnceGeometric':
        tfm = []
        for _ in range(self.B):
            p_rot = math.copysign(1, random.random() - (1 - p))
            tfm.append([[1, 0., 0], [0, p_rot, 0], [0, 0, 1]])
        tfm_matrix = torch.tensor(tfm)
        self._mix(tfm_matrix, prob)
        return self

    def apply(self, x: torch.Tensor) -> torch.Tensor:
        grid = F.affine_grid(self.m[:, :2, :], cast(List[int], x.shape))
        grid = grid.to(x.device)
        return F.grid_sample(x, grid, mode='bilinear', padding_mode='reflection')


class AllAtOnceColor:
    """
    Similar to AllAtOnceGeometric, performs multiple color transforms at once.

    Args:
        B (int): batch size
        init (torch.Tensor): an initial user supplied transformation matrix. If
            not provided, default to identity.
    """
    B: int
    m: torch.Tensor

    def __init__(self, B: int, init: Optional[torch.Tensor] = None):
        self.B = B
        if init is None:
            self.m = torch.stack([torch.eye(4, 4) for _ in range(B)], dim=0)
        else:
            self.m = init

    def _mix(self, m: torch.Tensor, prob: float) -> None:
        RND = (self.B, 1, 1)
        self.m = torch.where(
            torch.rand(RND) < prob, torch.bmm(m, self.m), self.m)

    def brightness(self, alpha: float, prob: float = 1.) -> 'AllAtOnceColor':
        """
        Change brightness by a factor alpha

        Args:
            alpha (float): scale factor
        """
        tfm = torch.stack([
            torch.eye(4, 4) * random.uniform(1 - alpha, 1 + alpha)
            for _ in range(self.B)
        ], dim=0)
        tfm[:, 3, 3] = 1
        self._mix(tfm, prob)
        return self

    def contrast(self, alpha: float, prob: float = 1.) -> 'AllAtOnceColor':
        """
        Scale contrast by factor alpha

        Args:
            alpha (float): scale factor
        """
        rot = []
        for _ in range(self.B):
            t = random.uniform(1 - alpha, 1 + alpha)
            rot.append([[t, 0, 0, (1 - t) / 2], [0, t, 0, (1 - t) / 2],
                        [0, 0, t, (1 - t) / 2], [0, 0, 0, 1]])
        rot_matrix = torch.tensor(rot)
        self._mix(rot_matrix, prob)
        return self

    def apply(self, x: torch.Tensor) -> torch.Tensor:
        """
        Applies transforms on x

        Args:
            x (torch.Tensor): input

        Returns:
            transformed x
        """
        B, C, H, W = x.shape
        w = self.m[:, :3, :3].reshape(3 * self.B, 3, 1, 1)
        b = self.m[:, :3, 3].reshape(self.B, 3, 1, 1)
        x = x.view(1, B * C, H, W)
        w = w.to(x.device)
        b = b.to(x.device)
        out = F.conv2d(x, w, None, groups=self.B)
        out = out.view(B, C, H, W)
        return out + b



import torch
from typing import List

import torchelie.nn as tnn
import torchelie.utils as tu
import torch.nn as nn


class Pix2PixHDGlobalGenerator(tnn.CondSeq):
    """
    Residual generator used in `Pix2PixHD <https://arxiv.org/abs/1711.11585>`_
    .

    :code:`arch` is a list of strings representing blocks.

    For example, this creates a first conv with 32 output channels, 3
    downsampling stride 2 convs that double the number of channels, 5 residual
    blocks, 3 upsampling convs halving the number of channels, and a final conv
    that converts back to RGB.

    :code:```
    Pix2PixHDGlobalGenerator(['32', 'd128', 'd512', 'd512', 'R512', 'R512',
        'R512', 'R512', 'R512', 'u512', 'u512', 'u128'])
    ```

    """

    def __init__(self, arch: List[str]) -> None:
        super().__init__()
        self.arch = arch
        self.to_standard_arch()

    def to_standard_arch(self):
        self._modules.clear()
        arch = self.arch
        self.input = tnn.ConvBlock(3, int(arch[0]), 7)
        ch, i = int(arch[0]), 1

        ii = 0
        self.encode = tnn.CondSeq()
        while arch[i][0] == 'd':
            out_ch = int(arch[i][1:])
            self.encode.add_module(f'conv_{ii}',
                                   tnn.ConvBlock(ch, out_ch, 3, stride=2))
            ch = out_ch
            i += 1
            ii += 1

        ii = 0
        self.transform = tnn.CondSeq()
        while arch[i][0] == 'R':
            out_ch = int(arch[i][1:])
            self.transform.add_module(f'transform_{ii}',
                                      tnn.PreactResBlock(ch, out_ch))
            ch = out_ch
            i += 1
            ii += 1

        ii = 0
        self.decode = tnn.CondSeq()
        while i < len(arch) and arch[i][0] == 'u':
            out_ch = int(arch[i][1:])
            self.decode.add_module(
                f'out_conv_{ii}',
                tnn.ConvBlock(ch, out_ch, 3, stride=1).add_upsampling())
            ch = out_ch
            i += 1
            ii += 1
        self.to_rgb = tnn.ConvBlock(out_ch, 3, 7).remove_batchnorm()
        self.to_rgb.relu = nn.Sigmoid()

        def to_instance_norm(m):
            if isinstance(m, nn.BatchNorm2d):
                return nn.InstanceNorm2d(m.num_features, affine=True)
            if isinstance(m, nn.Conv2d):
                m.padding_mode = 'reflect'
            return m

        tnn.utils.edit_model(self, to_instance_norm)

    def leaky(self) -> 'Pix2PixHDGlobalGenerator':

        def to_leaky(m):
            if isinstance(m, nn.ReLU):
                return nn.LeakyReLU(0.2, m.inplace)
            return m

        tnn.utils.edit_model(self, to_leaky)
        return self

    def to_unet(self) -> 'Pix2PixHDGlobalGenerator':
        self._modules.clear()
        arch = self.arch

        ch = int(self.arch[0])

        self.input = tnn.ConvBlock(33, int(arch[0]), 7)

        def _build(i, prev_ch):
            ch = int(arch[i][1:])

            if arch[i][0] == 'R':
                transforms = tnn.CondSeq()
                transforms.in_channels = prev_ch
                ii = 0
                while arch[i][0] == 'R':
                    ch = int(arch[i][1:])
                    transforms.add_module(f'transform_{ii}',
                                          tnn.PreactResBlock(prev_ch, ch))
                    prev_ch = ch
                    i += 1
                    ii += 1
                transforms.out_channels = ch
                return transforms
            if arch[i][0] == 'd':
                u = tnn.encdec.UBlock(prev_ch, ch, _build(i + 1, ch))
                u.to_bilinear_sampling()
                u.set_decoder_num_layers(1).set_encoder_num_layers(1)
                return u

        self.encdec = _build(1, ch)
        self.to_rgb = tnn.ConvBlock(ch, 3, 3).remove_batchnorm()
        self.to_rgb.relu = nn.Sigmoid()

        def to_instance_norm(m):
            if isinstance(m, nn.BatchNorm2d):
                return nn.InstanceNorm2d(m.num_features, affine=True)
            if isinstance(m, nn.Conv2d):
                m.padding_mode = 'reflect'
            return m

        tnn.utils.edit_model(self, to_instance_norm)

        return self

    def to_equal_lr(self, leak=0.) -> 'Pix2PixHDGlobalGenerator':
        return tnn.utils.net_to_equal_lr(self, leak=leak, mode='fan_in')


@tu.experimental
def pix2pix_res() -> Pix2PixHDGlobalGenerator:
    return Pix2PixHDGlobalGenerator(['64', 'd128', 'd512', 'd1024'] +
                                    ['R1024'] * 10 + ['u1024', 'u512', 'u128'])


@tu.experimental
def pix2pix_res_dev() -> Pix2PixHDGlobalGenerator:
    return Pix2PixHDGlobalGenerator(['8', 'd16', 'd32', 'd128', 'd256'] +
                                    ['R256'] * 10 +
                                    ['u256', 'u128', 'u32', 'u16'])



"""
Train a classifier.

It logs train and test loss and accuracy. It also displays confusion matrix and
gradient based feature visualization throughout training.

If you have an image training directory with a folder per class and a testing
directory with a folder per class, you can run it with a command line.

`python3 -m torchelie.recipes.classification --trainset path/to/train --testset
path/to/test`
"""
import torch
import torchvision.models as tvmodels

import torchelie.callbacks as tcb
import torchelie.utils as tu
from torchelie.recipes.trainandtest import TrainAndTest
from torchelie.optim import RAdamW


def Classification(model,
                   train_fun,
                   test_fun,
                   train_loader,
                   test_loader,
                   classes,
                   *,
                   visdom_env=None,
                   checkpoint=None,
                   test_every=1000,
                   log_every=100):
    """
    Classification training and testing loop. Both forward functions must
    return a per-batch loss and logits predictions. It expands from
    :code:`TrainAndTest`. Both :code:`train_fun` and :code:`test_fun` must
    :code:`return {'loss': batch_loss, 'preds': logits_predictions}`. The model
    is automatically registered and checkpointed as :code:`checkpoint['model']`,
    and put in eval mode when testing. The list of classes is checkpointed as
    well in :code:`checkpoint['classes']`.


    Training callbacks:

    - AccAvg for displaying accuracy
    - WindowedMetricAvg for displaying loss
    - ConfusionMatrix if len(classes) <= 25
    - ClassificationInspector
    - MetricsTable

    Inherited training callbacks:

    - Counter for counting iterations, connected to the testing loop as well
    - VisdomLogger
    - StdoutLogger

    Testing:

    Testing loop is in :code:`.test_loop`.

    Testing callbacks:

    - AccAvg
    - WindowedMetricAvg
    - ConfusionMatrix if :code:`len(classes) <= 25`
    - ClassificationInspector
    - MetricsTable

    Inherited testing callbacks:

    - VisdomLogger
    - StdoutLogger
    - Checkpoint saving the best testing loss


    Args:
        model (nn.Model): a model
        train_fun (Callabble): a function that takes a batch as a single
            argument, performs a training forward pass and return a dict of
            values to populate the recipe's state. It expects the logits
            predictions under key "preds", and this batch's loss under key
            "loss".
        test_fun (Callable): a function taking a batch as a single argument
            then performs something to evaluate your model and returns a dict
            to populate the state. It expects the logits
            predictions under key "preds", and this batch's loss under key
            "loss".
        train_loader (DataLoader): Training set dataloader
        test_loader (DataLoader): Testing set dataloader
        classes (list of str): classes name, in order
        visdom_env (str): name of the visdom environment to use, or None for
            not using Visdom (default: None)
        test_every (int): testing frequency, in number of iterations (default:
            1000)
        log_every (int): logging frequency, in number of iterations (default:
            100)
    """

    key_best = (lambda state: -state['test_loop']['callbacks']['state']
                ['metrics']['loss'])

    loop = TrainAndTest(model,
                        train_fun,
                        test_fun,
                        train_loader,
                        test_loader,
                        visdom_env=visdom_env,
                        test_every=test_every,
                        log_every=log_every,
                        checkpoint=checkpoint,
                        key_best=key_best)

    loop.callbacks.add_callbacks([
        tcb.AccAvg(),
        tcb.WindowedMetricAvg('loss'),
    ])
    loop.register('classes', classes)

    loop.test_loop.callbacks.add_callbacks([
        tcb.AccAvg(post_each_batch=False, avg_type='running'),
        tcb.WindowedMetricAvg('loss', False),
    ])

    if visdom_env is not None:
        if len(classes) <= 50:
            loop.callbacks.add_epilogues([
                tcb.ConfusionMatrix(classes),
            ])
        loop.callbacks.add_epilogues([
            tcb.ClassificationInspector(30, classes),
            tcb.MetricsTable()
        ])

    if len(classes) <= 50:
        loop.test_loop.callbacks.add_callbacks([
            tcb.ConfusionMatrix(classes),
        ])

    loop.test_loop.callbacks.add_callbacks([
        tcb.ClassificationInspector(30, classes, False),
        tcb.MetricsTable(False)
    ])
    return loop


def CrossEntropyClassification(model,
                               train_loader,
                               test_loader,
                               classes,
                               lr=3e-3,
                               beta1=0.9,
                               wd=1e-2,
                               visdom_env='main',
                               test_every=1000,
                               log_every=100,
                               checkpoint='model'):
    """
    Extends Classification with default cross entropy forward passes. Also adds
    RAdamW and ReduceLROnPlateau.

    Inherited training callbacks:

    - AccAvg for displaying accuracy
    - WindowedMetricAvg for displaying loss
    - ConfusionMatrix if len(classes) <= 25
    - ClassificationInspector
    - MetricsTable
    - Counter for counting iterations, connected to the testing loop as well
    - VisdomLogger
    - StdoutLogger

    Training callbacks:

    - Optimizer with RAdamW
    - LRSched with ReduceLROnPlateau

    Testing:

    Testing loop is in :code:`.test_loop`.

    Inherited testing callbacks:

    - AccAvg
    - WindowedMetricAvg
    - ConfusionMatrix if :code:`len(classes) <= 25`
    - ClassificationInspector
    - MetricsTable
    - VisdomLogger
    - StdoutLogger
    - Checkpoint saving the best testing loss

    Args:
        model (nn.Module): a model learnable with cross entropy
        train_loader (DataLoader): Training set dataloader
        test_loader (DataLoader): Testing set dataloader
        classes (list of str): classes name, in order
        lr (float): the learning rate
        beta1 (float): RAdamW's beta1
        wd (float): weight decay
        visdom_env (str): name of the visdom environment to use, or None for
            not using Visdom (default: None)
        test_every (int): testing frequency, in number of iterations (default:
            1000)
        log_every (int): logging frequency, in number of iterations (default:
            1000)
    """

    def train_step(batch):
        x, y = batch
        pred = model(x)
        loss = torch.nn.functional.cross_entropy(pred, y)
        loss.backward()
        return {'loss': loss, 'pred': pred}

    def validation_step(batch):
        x, y = batch
        pred = model(x)
        loss = torch.nn.functional.cross_entropy(pred, y)
        return {'loss': loss, 'pred': pred}

    loop = Classification(model,
                          train_step,
                          validation_step,
                          train_loader,
                          test_loader,
                          classes,
                          visdom_env=visdom_env,
                          test_every=test_every,
                          log_every=log_every,
                          checkpoint=checkpoint)

    opt = RAdamW(model.parameters(),
                 lr=lr,
                 betas=(beta1, 0.999),
                 weight_decay=wd)
    loop.callbacks.add_callbacks([
        tcb.Optimizer(opt, log_lr=True),
        tcb.LRSched(torch.optim.lr_scheduler.ReduceLROnPlateau(opt))
    ])
    return loop


def MixupClassification(model,
                        train_loader,
                        test_loader,
                        classes,
                        *,
                        lr=3e-3,
                        beta1=0.9,
                        wd=1e-2,
                        visdom_env='main',
                        test_every=1000,
                        log_every=100):
    """
    A Classification recipe with a default froward training / testing pass
    using cross entropy and mixup, and extended with RAdamW and
    ReduceLROnPlateau.

    Args:
        model (nn.Module): a model learnable with cross entropy
        train_loader (DataLoader): Training set dataloader. Must have soft
            targets. Should be a DataLoader loading a MixupDataset or
            compatible.
        test_loader (DataLoader): Testing set dataloader. Dataset must have
            categorical targets.
        classes (list of str): classes name, in order
        lr (float): the learning rate
        beta1 (float): RAdamW's beta1
        wd (float): weight decay
        visdom_env (str): name of the visdom environment to use, or None for
            not using Visdom (default: None)
        test_every (int): testing frequency, in number of iterations (default:
            1000)
        log_every (int): logging frequency, in number of iterations (default:
            1000)
    """

    from torchelie.loss import continuous_cross_entropy

    def train_step(batch):
        x, y = batch
        pred = model(x)
        loss = continuous_cross_entropy(pred, y)
        loss.backward()
        return {'loss': loss}

    def validation_step(batch):
        x, y = batch
        pred = model(x)
        loss = torch.nn.functional.cross_entropy(pred, y)
        return {'loss': loss, 'pred': pred}

    loop = TrainAndTest(model,
                        train_step,
                        validation_step,
                        train_loader,
                        test_loader,
                        visdom_env=visdom_env,
                        test_every=test_every,
                        log_every=log_every)

    loop.callbacks.add_callbacks([
        tcb.WindowedMetricAvg('loss'),
    ])
    loop.register('classes', classes)

    loop.test_loop.callbacks.add_callbacks([
        tcb.AccAvg(post_each_batch=False),
        tcb.WindowedMetricAvg('loss', False),
    ])

    if visdom_env is not None:
        loop.callbacks.add_epilogues([
            tcb.MetricsTable()
        ])

    if len(classes) <= 25:
        loop.test_loop.callbacks.add_callbacks([
            tcb.ConfusionMatrix(classes),
        ])

    loop.test_loop.callbacks.add_callbacks([
        tcb.ClassificationInspector(30, classes, False),
        tcb.MetricsTable(False)
    ])

    opt = RAdamW(model.parameters(),
                 lr=lr,
                 betas=(beta1, 0.999),
                 weight_decay=wd)
    loop.callbacks.add_callbacks([
        tcb.Optimizer(opt, log_lr=True),
        tcb.LRSched(torch.optim.lr_scheduler.ReduceLROnPlateau(opt))
    ])
    return loop


if __name__ == '__main__':
    import argparse

    from torchelie.datasets import CachedDataset
    from torchvision.datasets import ImageFolder
    from torch.utils.data import DataLoader
    import torchvision.transforms as TF
    import torchelie.transforms as TTF

    parser = argparse.ArgumentParser()
    parser.add_argument('--trainset', type=str, required=True)
    parser.add_argument('--testset', type=str, required=True)
    parser.add_argument('--batch-size', type=int, default=32)
    parser.add_argument('--lr', type=float, default=0.01)
    parser.add_argument('--beta1', type=float, default=0.9)
    parser.add_argument('--wd', type=float, default=1e-2)
    parser.add_argument('--im-size', type=int, default=64)
    parser.add_argument('--device', type=str, default='cuda')
    parser.add_argument('--visdom-env', type=str)
    parser.add_argument('--no-cache', action='store_false')
    parser.add_argument('--epochs', type=int, default=5)
    parser.add_argument('--mixup', action='store_true')
    args = parser.parse_args()

    tfm = TF.Compose([
        TF.RandomResizedCrop(args.im_size, (0.7, 1.1)),
        TF.ColorJitter(0.5, 0.5, 0.4, 0.05),
        TF.RandomHorizontalFlip(),
        TF.ToTensor(),
        TF.Normalize([0.5] * 3, [0.5] * 3, True),
    ])
    tfm_test = TF.Compose([
        TTF.ResizedCrop(args.im_size, scale=1),
        TF.Resize(args.im_size),
        TF.ToTensor(),
        TF.Normalize([0.5] * 3, [0.5] * 3, True)
    ])

    if args.no_cache:
        trainset = ImageFolder(args.trainset, transform=tfm)
        testset = ImageFolder(args.testset, transform=tfm_test)
    else:
        trainset = ImageFolder(args.trainset)
        testset = ImageFolder(args.testset)

        trainset = CachedDataset(trainset, transform=tfm)
        testset = CachedDataset(testset, transform=tfm_test)

    if args.mixup:
        from torchelie.datasets import MixUpDataset
        trainset = MixUpDataset(trainset)

    trainloader = DataLoader(trainset,
                             args.batch_size,
                             num_workers=8,
                             pin_memory=True,
                             shuffle=True,
                             drop_last=True)
    testloader = DataLoader(testset,
                            args.batch_size,
                            num_workers=8,
                            pin_memory=True)

    model = tvmodels.resnet18(pretrained=False)
    model.fc = tu.kaiming(torch.nn.Linear(512, len(testset.classes)))

    if args.mixup:
        clf_recipe = MixupClassification(model,
                                         trainloader,
                                         testloader,
                                         testset.classes,
                                         log_every=10,
                                         test_every=50,
                                         lr=args.lr,
                                         beta1=args.beta1,
                                         wd=args.wd,
                                         visdom_env=args.visdom_env)
    else:
        clf_recipe = CrossEntropyClassification(model,
                                                trainloader,
                                                testloader,
                                                testset.classes,
                                                log_every=10,
                                                test_every=50,
                                                lr=args.lr,
                                                beta1=args.beta1,
                                                wd=args.wd,
                                                visdom_env=args.visdom_env)

    clf_recipe.to(args.device)
    clf_recipe.run(args.epochs)



"""
SUPER ALPHA
"""

import random
import os

from typing import List
import PIL


class ImageLogger:
    """
    Logs images and decisions up to a capacity per class
    """
    def __init__(self, root: str, classes: List[str], capacity: int = 200):
        self.root = root
        self.classes = classes
        self.capacity = capacity
        for kls in classes:
            os.makedirs(root + '/' + kls, exist_ok=True)

    def __call__(self, images, klass: List[int]) -> None:
        for i, k in zip(images, klass):
            k_name = self.classes[k]
            n = random.randint(-int(self.capacity * 0.25), self.capacity)
            if n < 0:
                continue
            path = f'{self.root}/{k_name}/{n}.jpg'
            if isinstance(i, PIL.Image.Image):
                i.save(path)



"""
Recipes are a way to provide off the shelf algorithms that ce be either used
directly from the command line or easily imported in a python script if more
flexibility is needed.

A recipe should, as much a possible, be agnostic of the data and the underlying
model so that it can be used as a way to quickly try an algorithm on new data
or be easily experimented on by changing the model
"""
from torchelie.recipes.recipebase import Recipe
from torchelie.recipes.classification import Classification
from torchelie.recipes.classification import CrossEntropyClassification
from torchelie.recipes.classification import MixupClassification
from torchelie.recipes.deepdream import DeepDream
from torchelie.recipes.feature_vis import FeatureVis
from torchelie.recipes.neural_style import NeuralStyle
from torchelie.recipes.trainandtest import TrainAndTest
from torchelie.recipes.trainandcall import TrainAndCall
from torchelie.recipes.gan import GANRecipe



import math
import torch
from torch.optim import Optimizer
from collections import defaultdict


class DeepDreamOptim(Optimizer):
    r"""Optimizer used by Deep Dream. It rescales the gradient by the average of
    the absolute values of the gradient.

    :math:`\theta_i := \theta_i - lr \frac{g_i}{\epsilon+\frac{1}{M}\sum_j^M |g_j|}`

    Args:
        params: parameters as expected by Pytorch's optimizers
        lr (float): the learning rate
        eps (float): epsilon value to avoid dividing by zero
    """
    def __init__(self, params, lr=1e-3, eps=1e-8, weight_decay=0):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        defaults = dict(lr=lr, eps=eps, weight_decay=weight_decay)
        super(DeepDreamOptim, self).__init__(params, defaults)

    def step(self, closure=None):
        """Update the weights

        Args:
            closure (optional fn): a function that computes gradients
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                # Perform stepweight decay
                p.data.mul_(1 - group['lr'] * group['weight_decay'])

                step_size = group['lr']
                eps = group['eps']

                p.data.add_(p.grad.data,
                            alpha=-step_size
                            / (eps + p.grad.data.abs().mean()))

        return loss


class AddSign(Optimizer):
    r"""AddSign optimizer from Neural Optimiser search with Reinforcment
    learning (Bello et al, 2017)

    :math:`\theta_i := \theta_i - \text{lr}(1+\text{sign}(g)*\text{sign}(m))*g`

    Args:
        params: parameters as expected by Pytorch's optimizers
        lr (float): the learning rate
        eps (float): epsilon value to avoid dividing by zero
    """
    def __init__(self, params, lr=1e-3, beta=0.9, weight_decay=0):
        defaults = {'lr': lr, 'beta': beta, 'weight_decay': weight_decay}
        super(AddSign, self).__init__(params, defaults)

    def step(self, closure=None):
        """Update the weights

        Args:
            closure (optional fn): a function that computes gradients
        """
        loss = None

        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['mom'] = torch.zeros_like(p.data)

                lr = group['lr']
                beta = group['beta']
                weight_decay = group['weight_decay']

                state['step'] += 1
                mom = state['mom']
                mom.mul_(beta).add_(grad, alpha=1 - beta)

                s = torch.sign(grad).mul_(torch.sign(mom)).add_(1)

                p.data.addcmul_(-lr * s, grad)
                p.data.add_(p.data, alpha=-lr * weight_decay)

        return loss


class RAdamW(Optimizer):
    r"""Implements RAdamW algorithm.

    RAdam from `On the Variance of the Adaptive Learning Rate and Beyond
    <https://arxiv.org/abs/1908.03265v1>`_

    * `Adam: A Method for Stochastic Optimization
      <https://arxiv.org/abs/1412.6980>`_
    * `Decoupled Weight Decay Regularization
      <https://arxiv.org/abs/1711.05101>`_
    * `On the Convergence of Adam and Beyond
      <https://openreview.net/forum?id=ryQu7f-RZ>`_
    * `On the Variance of the Adaptive Learning Rate and Beyond
      <https://arxiv.org/abs/1908.03265v1>`_

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay coefficient (default: 1e-2)
    """
    def __init__(self,
                 params,
                 lr=1e-3,
                 betas=(0.9, 0.999),
                 eps=1e-8,
                 weight_decay=1e-2,
                 stable=False):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(
                betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(
                betas[1]))
        defaults = dict(lr=lr,
                        betas=betas,
                        eps=eps,
                        weight_decay=weight_decay,
                        stable=stable)
        super(RAdamW, self).__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                # Perform optimization step
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        'Adam does not support sparse gradients, please consider SparseAdam instead'
                    )

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                eps = group['eps']
                lr = group['lr']
                stable = group['stable']
                if 'rho_inf' not in group:
                    group['rho_inf'] = 2 / (1 - beta2) - 1
                rho_inf = group['rho_inf']

                state['step'] += 1
                t = state['step']

                # Decay the first and second moment running average coefficient
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_no_bias = exp_avg / (1 - beta1**t)
                rho_t = rho_inf - ((2 * t * (beta2**t)) / (1 - beta2**t))
                var = exp_avg_sq / (1 - beta2**t)

                # Perform stepweight decay
                if stable:
                    # AdamS
                    p.data.mul_(1 - group['lr'] * group['weight_decay']
                                / var.mean().add_(1e-8).sqrt_())
                else:
                    # AdamW
                    p.data.mul_(1 - group['lr'] * group['weight_decay'])

                if rho_t > 4:
                    var.sqrt_()
                    r = math.sqrt(((rho_t - 4) * (rho_t - 2) * rho_inf)
                                  / ((rho_inf - 4) * (rho_inf - 2) * rho_t))

                    p.data.addcdiv_(exp_avg_no_bias, var + eps, value=-lr * r)
                else:
                    p.data.add_(exp_avg_no_bias, alpha=-lr)

        return loss


class Lookahead(Optimizer):
    """
    Implements Lookahead from `Lookahead Optimizer: k steps forward, 1 step
    back` (Zhang et al, 2019)

    Args:
        base_optimizer (Optimizer): an optimizer for the inner loop
        alpha (float): outer loop learning rate
        k (int): number of steps in the inner loop
    """
    def __init__(self, base_optimizer, alpha=0.5, k=5):
        if not 0.0 <= alpha <= 1.0:
            raise ValueError('Invalid slow update rate: ' + str(alpha))
        if not 1 <= k:
            raise ValueError('Invalid lookahead steps: ' + str(k))

        self.alpha = alpha
        self.k = k
        self.optimizer = base_optimizer
        self.state: dict = defaultdict(dict)
        self.param_groups = base_optimizer.param_groups

    def state_dict(self):
        return {'lookahead': self.state, 'opt': self.optimizer.state_dict()}

    def load_state_dict(self, state):
        self.state = state['lookahead']
        self.optimizer.load_state_dict(state['opt'])

    def zero_grad(self):
        self.optimizer.zero_grad()

    def step(self, closure=None):
        """Performs a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()
        loss = self.optimizer.step()
        if 'steps' not in self.state:
            self.state['steps'] = 0
        steps = self.state['steps']

        self.state['steps'] += 1

        if steps % self.k != 0:
            return loss

        for group in self.optimizer.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                if p not in self.state:
                    self.state[p] = p.detach().clone()
                    self.state[p].requires_grad = False

                q = self.state[p]
                q.data.add_(p.data - q.data, alpha=self.alpha)
                p.data.copy_(q.data)
        return loss

    def __repr__(self):
        return "Lookahead({}, alpha={}, k={})".format(repr(self.optimizer),
                                                      self.alpha, self.k)



"""
Deep Image Prior as a command line tool is in this recipe, it provides
super-resolution and denoising.

:code:`python3 -m torchelie.recipes.image_prior` for the usage message
"""
import torch
import torch.nn.functional as F
from PIL import Image
import torchvision.transforms.functional as TFF
from torchelie.models import Hourglass
import torchelie.transforms as ttf
import torchelie as tch
from torchelie.recipes.trainandcall import TrainAndCall
import torchelie.callbacks as tcb


def with_patches(img, patch_size, task, *args, **kwargs):
    if img.width < patch_size and img.height < patch_size:
        return task(img, *args, **kwargs)

    transformed = [(coords, task(im, *args, **kwargs))
                   for coords, im in ttf.patches(img, patch_size)]

    return ttf.paste_patches(transformed)


def with_patches2(img, mask, patch_size, task, *args, **kwargs):
    assert img.size == mask.size, 'Sizes of mask and image do not match'

    if img.width < patch_size and img.height < patch_size:
        return task(img, mask, *args, **kwargs)

    transformed = [(coords, task(im, ma, *args, **kwargs))
                   for (coords,
                        im), (_, ma) in zip(ttf.patches(img, patch_size),
                                            ttf.patches(mask, patch_size))]

    return ttf.paste_patches(transformed)


def input_noise(size, channels):
    if channels != 2:
        lines = torch.linspace(-0.01, 0.01, size[0])
        cols = torch.linspace(-0.01, 0.01, size[1])
        ll, cc = torch.meshgrid(lines, cols)
        ll = ll.unsqueeze(0).unsqueeze(0)
        cc = cc.unsqueeze(0).unsqueeze(0)
        return cc + ll + torch.rand(1, channels, size[0], size[1]) / 10 - 0.05
    else:
        lines = torch.linspace(-4, 4, size[1])
        cols = torch.linspace(-4, 4, size[0])
        ll, cc = torch.meshgrid(lines, cols)
        ll = ll.unsqueeze(0).unsqueeze(0)
        cc = cc.unsqueeze(0).unsqueeze(0)
        return torch.cat([ll, cc], dim=1)


def inpainting(img,
               mask,
               hourglass,
               input_dim,
               iters,
               lr,
               noise_std=1 / 30,
               device='cuda'):
    im = TFF.to_tensor(img)[None].to(device)
    mask = TFF.to_tensor(mask)[None].to(device)
    z = input_noise((im.shape[2], im.shape[3]), input_dim)
    z = z.to(device)
    print(hourglass)

    def body(batch):
        recon = hourglass(z + torch.randn_like(z) * noise_std)
        loss = torch.sum(
            F.mse_loss(F.interpolate(recon, size=im.shape[2:], mode='nearest'),
                       im,
                       reduction='none') * mask / mask.sum())
        loss.backward()
        return {"loss": loss}

    def display():
        recon = hourglass(z)
        recon = F.interpolate(recon, size=im.shape[2:], mode='nearest')
        loss = F.mse_loss(recon * mask, im)

        result = recon * (1 - mask) + im * mask
        return {
            "loss": loss,
            "recon": recon.clamp(0, 1),
            'orig': im,
            'result': result.clamp(0, 1)
        }

    loop = make_loop(hourglass, body, display, iters, lr)
    loop.test_loop.callbacks.add_callbacks([tcb.Log('result', 'result')])
    loop.to(device)
    loop.run(1)
    with torch.no_grad():
        hourglass.eval()
        return TFF.to_pil_image(hourglass(z)[0].cpu())


def superres(img,
             hourglass,
             input_dim,
             scale,
             iters,
             lr,
             noise_std=1 / 30,
             device='cuda'):
    im = TFF.to_tensor(img)[None].to(device)
    z = input_noise((im.shape[2] * scale, im.shape[3] * scale), input_dim)
    z = z.to(device)

    def body(batch):
        recon = hourglass(z + torch.randn_like(z) * noise_std)
        loss = F.mse_loss(
            F.interpolate(recon, size=im.shape[2:], mode='bilinear'), im)
        loss.backward()
        return {
            "loss": loss,
        }

    def display():
        recon = hourglass(z)
        loss = F.mse_loss(
            F.interpolate(recon, size=im.shape[2:], mode='bilinear'), im)
        return {
            "loss":
            loss,
            "recon":
            recon.clamp(0, 1),
            'orig':
            F.interpolate(im, scale_factor=scale, mode='bicubic').clamp(0, 1)
        }

    loop = make_loop(hourglass, body, display, iters, lr)
    loop.to(device)
    loop.run(1)
    with torch.no_grad():
        hourglass.eval()
        return TFF.to_pil_image(hourglass(z)[0].cpu())


def make_loop(hourglass, body, display, num_iter, lr):
    loop = TrainAndCall(hourglass,
                        body,
                        display,
                        range(num_iter),
                        test_every=50,
                        checkpoint=None)
    opt = tch.optim.RAdamW(hourglass.parameters(), lr=lr)
    loop.callbacks.add_callbacks([
        tcb.WindowedMetricAvg('loss'),
        tcb.Optimizer(opt, clip_grad_norm=0.5, log_lr=True),
    ])
    loop.test_loop.callbacks.add_callbacks([
        tcb.Log('recon', 'img'),
        tcb.Log('orig', 'orig'),
        tcb.Log('loss', 'loss'),
    ])
    return loop


if __name__ == '__main__':
    import argparse

    hyper_params = {
        '2xsuperres': {
            'fun': 'superres',
            'noise_std': 1 / 30,
            'lr': 0.01,
            'iters': 2000,
            'model': 'superres',
            'scale': 2
        },
        '4xsuperres': {
            'fun': 'superres',
            'noise_std': 1 / 30,
            'lr': 0.01,
            'iters': 2000,
            'model': 'superres',
            'scale': 4
        },
        '8xsuperres': {
            'fun': 'superres',
            'noise_std': 1 / 20,
            'lr': 0.01,
            'iters': 4000,
            'model': 'superres',
            'scale': 8
        },
        'text-inpainting': {
            'fun': 'inpainting',
            'noise_std': 1 / 30,
            'lr': 0.01,
            'iters': 6000,
            'model': 'superres'
        },
        'hole-inpainting-1': {
            'fun': 'inpainting',
            'noise_std': 1 / 30,
            'lr': 0.01,
            'iters': 6000,
            'model': 'inpainting-1'
        },
        'hole-inpainting-2': {
            'fun': 'inpainting',
            'noise_std': 0,
            'lr': 0.1,
            'iters': 5000,
            'model': 'inpainting-2'
        },
        'denoising': {
            'fun': 'superres',
            'noise_std': 1 / 30,
            'lr': 0.01,
            'iters': 1800,
            'model': 'superres',
            'scale': 1
        },
        'jpeg-removal': {
            'fun': 'superres',
            'noise_std': 1 / 30,
            'lr': 0.01,
            'iters': 2400,
            'model': 'inpainting-2',
            'scale': 1
        }
    }
    models = {
        'superres': {
            'noise_dim': 32,
            'down_channels': [128] * 6,
            'down_kernel': [3] * 6,
            'up_kernel': [3] * 6,
            'skip_channels': 4,
            'upsampling': 'bilinear'
        },
        'inpainting-1': {
            'noise_dim': 2,
            'down_channels': [128] * 6,
            'down_kernel': [3] * 6,
            'up_kernel': [3] * 6,
            'skip_channels': 0,
            'upsampling': 'bilinear'
        },
        'inpainting-2': {
            'noise_dim': 32,
            'down_channels': [16, 32, 64, 128, 128, 128],
            'down_kernel': [3] * 6,
            'up_kernel': [5] * 6,
            'skip_channels': 0,
            'upsampling': 'nearest'
        }
    }

    parser = argparse.ArgumentParser()
    parser.add_argument('--input', required=True, type=str)
    parser.add_argument('--task',
                        required=True,
                        choices=hyper_params.keys(),
                        type=str)
    parser.add_argument('--mask', type=str)
    parser.add_argument('--lr', type=float)
    parser.add_argument('--noise-std', type=float)
    parser.add_argument('--output', default='out.jpg', type=str)
    parser.add_argument('--n-layers', default=6, type=int)
    parser.add_argument('--device', default='cuda')
    parser.add_argument('--max-size', type=int)
    parser.add_argument('--patch-size', default=1024, type=int)
    opts = parser.parse_args()

    im = Image.open(opts.input).convert('RGB')
    if opts.max_size is not None:
        im.thumbnail((opts.max_size, opts.max_size))

    hps = hyper_params[opts.task]
    for k, v in hps.items():
        hps[k] = getattr(opts, k, v) or v

    model_params = models[hps.pop('model')]
    for k, v in model_params.items():
        if isinstance(v, list):
            model_params[k] = v[:opts.n_layers]

    model = Hourglass(**model_params)

    fun_name = hps.pop('fun')
    fun = globals()[fun_name]
    print(fun, hps)

    if 'inpainting' in fun_name:
        if opts.mask is not None:
            mask = Image.open(opts.mask).convert('RGB')
            if opts.max_size is not None:
                mask.thumbnail((opts.max_size, opts.max_size))

        out = with_patches2(im, mask, opts.patch_size, inpainting, model,
                            model_params['noise_dim'], **hps)
    else:
        out = with_patches(im, opts.patch_size, fun, model,
                           model_params['noise_dim'], **hps)

    out.save(opts.output)



"""
Feature visualization

This optimizes an image to maximize some neuron in order to visualize the
features it captures

A commandline is provided with `python3 -m torchelie.recipes.feature_vis`
"""
import random

import torch
import torchvision.transforms as TF

import torchelie.nn as tnn
import torchelie.callbacks as tcb
from torchelie.optim import DeepDreamOptim
from torchelie.recipes.recipebase import Recipe
from torchelie.data_learning import ParameterizedImg


class FeatureVis(torch.nn.Module):
    """
    Feature viz

    First instantiate the recipe then call `recipe(n_iter, img)`

    Args:
        model (nn.Module): the trained model to use
        layer (str): the layer to use on which activations will be maximized
        input_size (int, or (int, int)): the size of the image the model
            accepts as input
        lr (float, optional): the learning rate
        device (device): where to run the computation
        visdom_env (str or None): the name of the visdom env to use, or None
            to disable Visdom
    """
    def __init__(self,
                 model,
                 layer,
                 input_size,
                 lr=1e-3,
                 device='cpu',
                 visdom_env='feature_vis'):
        super().__init__()
        self.device = device
        self.model = tnn.WithSavedActivations(model, names=[layer])
        self.layer = layer
        if isinstance(input_size, (list, tuple)):
            self.input_size = input_size
        else:
            self.input_size = (input_size, input_size)
        self.norm = tnn.ImageNetInputNorm()
        self.lr = lr
        self.visdom_env = visdom_env

    def fit(self, n_iters, neuron):
        """
        Run the recipe

        Args:
            n_iters (int): number of iterations to run
            neuron (int): the feature map to maximize

        Returns:
            the optimized image
        """
        canvas = ParameterizedImg(1, 3,
                                  self.input_size[0] + 10,
                                  self.input_size[1] + 10)

        def forward(_):
            cim = canvas()
            rnd = random.randint(0, cim.shape[2] // 10)
            im = cim[:, :, rnd:, rnd:]
            im = torch.nn.functional.interpolate(im,
                                                 size=(self.input_size[0],
                                                       self.input_size[1]),
                                                 mode='bilinear')
            _, acts = self.model(self.norm(im), detach=False)
            fmap = acts[self.layer]
            loss = -fmap[0][neuron].sum()
            loss.backward()

            return {'loss': loss, 'img': cim}

        loop = Recipe(forward, range(n_iters))
        loop.register('canvas', canvas)
        loop.register('model', self)
        loop.callbacks.add_callbacks([
            tcb.Counter(),
            tcb.Log('loss', 'loss'),
            tcb.Log('img', 'img'),
            tcb.Optimizer(DeepDreamOptim(canvas.parameters(), lr=self.lr)),
            tcb.VisdomLogger(visdom_env=self.visdom_env, log_every=10),
            tcb.StdoutLogger(log_every=10)
        ])
        loop.to(self.device)
        loop.run(1)
        return canvas.render().cpu()


if __name__ == '__main__':
    import argparse
    import torchvision.models as tvmodels

    models = {
        'vgg': {
            'ctor': tvmodels.vgg19,
            'sz': 224
        },
        'resnet': {
            'ctor': tvmodels.resnet18,
            'sz': 224
        },
        'googlenet': {
            'ctor': tvmodels.googlenet,
            'sz': 299
        },
        'inception': {
            'ctor': tvmodels.inception_v3,
            'sz': 299
        },
    }
    parser = argparse.ArgumentParser(description="Feature visualization")
    parser.add_argument('--model', required=True, choices=models.keys())
    parser.add_argument('--layer', required=True)
    parser.add_argument('--input-size')
    parser.add_argument('--neuron', required=True, type=int)
    parser.add_argument('--lr', type=float, default=1e-3)
    parser.add_argument('--out', default='features.png')
    parser.add_argument('--device', default='cuda')
    parser.add_argument('--iters', default=4000, type=int)
    parser.add_argument('--visdom-env')
    args = parser.parse_args()

    choice = models[args.model]
    model = choice['ctor'](pretrained=True)
    print(model)
    fv = FeatureVis(model,
                    args.layer,
                    args.input_size or choice['sz'],
                    lr=args.lr,
                    device=args.device,
                    visdom_env=args.visdom_env)
    out = fv.fit(args.iters, args.neuron)
    TF.ToPILImage()(out).save(args.out)



import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import List, Callable, cast

from torchelie.nn import ImageNetInputNorm
from torchelie.models import PerceptualNet


class PerceptualLoss(nn.Module):
    r"""
    Perceptual loss: the distance between a two images deep representation

    :math:`\text{Percept}(\text{input}, \text{target})=\sum_l^{layers}
    \text{loss_fn}(\text{Vgg}(\text{input})_l, \text{Vgg}(\text{target})_l)`

    Args:
        l (list of str): the layers on which to compare the representations
        rescale (bool): whether to scale images smaller side to 224 as
            expected by the underlying vgg net
        loss_fn (distance function): a distance function to compare the
            representations, like mse_loss or l1_loss
    """
    def __init__(self,
                 layers: List[str],
                 rescale: bool = False,
                 loss_fn: Callable[[torch.Tensor, torch.Tensor],
                                   torch.Tensor] = F.mse_loss,
                 use_avg_pool: bool = True,
                 remove_unused_layers: bool = True):
        super(PerceptualLoss, self).__init__()
        self.m = PerceptualNet(layers,
                               use_avg_pool=use_avg_pool,
                               remove_unused_layers=remove_unused_layers)
        self.norm = ImageNetInputNorm()
        self.rescale = rescale
        self.loss_fn = loss_fn

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Return the perceptual loss between batch of images `x` and `y`
        """
        if self.rescale:
            s = 224 / min(y.shape[-2:])
            y = F.interpolate(y, scale_factor=s, mode='bicubic')
            s = 224 / min(x.shape[-2:])
            x = F.interpolate(x, scale_factor=s, mode='bicubic')

        _, ref = self.m(self.norm(y), detach=True)
        _, acts = self.m(self.norm(x), detach=False)
        loss = cast(torch.Tensor,
                    sum(self.loss_fn(acts[k], ref[k]) for k in acts.keys()))
        return loss / len(acts)



r"""
Hinge loss from Spectral Normalization GAN.

https://arxiv.org/abs/1802.05957

:math:`L_D(x_r, x_f) = \text{max}(0, 1 - D(x_r)) + \text{max}(0, 1 + D(x_f))`

:math:`L_G(x_f) = -D(x_f)`
"""

import torch
import torch.nn.functional as F


def real(x: torch.Tensor, reduction: str = 'mean') -> torch.Tensor:
    out = F.relu(1 - x)
    if reduction == 'none':
        return out
    if reduction == 'mean':
        return out.mean()
    if reduction == 'sum':
        return out.sum()
    assert False, f'{reduction} is not a valid reduction method'


def fake(x: torch.Tensor, reduction: str = 'mean') -> torch.Tensor:
    out = F.relu(1 + x)
    if reduction == 'none':
        return out
    if reduction == 'mean':
        return out.mean()
    if reduction == 'sum':
        return out.sum()
    assert False, f'{reduction} is not a valid reduction method'


def generated(x: torch.Tensor, reduction: str = 'mean') -> torch.Tensor:
    out = -x
    if reduction == 'none':
        return out
    if reduction == 'mean':
        return out.mean()
    if reduction == 'sum':
        return out.sum()
    assert False, f'{reduction} is not a valid reduction method'



import torch
import torch.nn as nn
import torchelie.nn as tnn
import torchelie.utils as tu

from typing import List, Optional
from .classifier import ClassificationHead

PREACT_BLOCKS = (tnn.PreactResBlock, tnn.PreactResBlockBottleneck)
STD_BLOCKS = (tnn.ResBlock, tnn.ResBlockBottleneck)
BLOCKS = PREACT_BLOCKS + STD_BLOCKS
BOTTLENECKS_BLOCKS = (tnn.PreactResBlockBottleneck, tnn.ResBlockBottleneck)


class ResNetInput(nn.Module):
    relu: Optional[nn.ReLU]
    pool: Optional[nn.Module]

    def __init__(self, in_channels: int = 3, out_channels: int = 64) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.conv = tu.kaiming(
            tnn.Conv2d(in_channels, out_channels, 7, stride=2))
        self.relu = nn.ReLU(True)
        self.pool = nn.MaxPool2d(3, 2, 1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        if self.relu is not None:
            x = self.relu(x)
        if self.pool is not None:
            x = self.pool(x)
        return x

    def set_input_specs(self, input_size: int, in_channels=3) -> 'ResNetInput':
        self.in_channels = in_channels
        in_ch = self.in_channels
        out_ch = self.out_channels

        if input_size <= 64:
            self.conv = tu.kaiming(tnn.Conv2d(in_ch, out_ch, ks=3))
            del self.pool
            self.pool = None
        elif input_size <= 128:
            self.conv = tu.kaiming(tnn.Conv2d(in_ch, out_ch, 5, stride=2))
            del self.pool
            self.pool = None
        else:
            self.conv = tu.kaiming(tnn.Conv2d(in_ch, out_ch, 7, stride=2))
            if self.pool is None:
                self.pool = nn.MaxPool2d(3, 2, 1)

        return self


class ResNet(nn.Module):
    def __init__(self, arch: List[str], num_classes: int) -> None:
        super().__init__()

        def parse(layer: str) -> List[int]:
            return [int(x) for x in layer.split(':')]

        self.arch = list(map(parse, arch))

        self.features = tnn.CondSeq()
        self.features.add_module('input', ResNetInput(3, self.arch[0][0]))

        self._change_block_type('basic')
        self.classifier = ClassificationHead(self.arch[-1][0], num_classes)

    def _make_block(self, block_type: str, in_ch: int, out_ch: int,
                    stride: int) -> nn.Module:
        if block_type == 'basic':
            return tnn.ResBlock(in_ch, out_ch, stride)
        if block_type == 'bottleneck':
            return tnn.ResBlockBottleneck(in_ch, out_ch, stride)
        if block_type == 'preact_basic':
            return tnn.PreactResBlock(in_ch, out_ch, stride)
        if block_type == 'preact_bottleneck':
            return tnn.PreactResBlockBottleneck(in_ch, out_ch, stride)
        if block_type == 'resnext':
            return tnn.ResBlockBottleneck(in_ch, out_ch, stride).resnext()
        if block_type == 'preact_resnext':
            return tnn.PreactResBlockBottleneck(in_ch, out_ch,
                                                stride).resnext()
        if block_type == 'wide':
            return tnn.ResBlockBottleneck(in_ch, out_ch, stride).wide()
        if block_type == 'preact_wide':
            return tnn.PreactResBlockBottleneck(in_ch, out_ch, stride).wide()
        assert False

    def _change_block_type(self, ty: str) -> None:
        arch = self.arch

        feats = tnn.CondSeq()
        assert isinstance(self.features.input, ResNetInput)
        feats.add_module('input', self.features.input)
        in_ch = arch[0][0]
        self.in_channels = in_ch

        for i, (ch, s) in enumerate(arch):
            feats.add_module(f'block_{i}',
                             self._make_block(ty, in_ch, ch, stride=s))
            in_ch = ch
        self.out_channels = ch

        if 'preact' in ty:
            assert isinstance(feats.block_0, PREACT_BLOCKS)
            feats.block_0.no_preact()
            feats.add_module('final_bn', nn.BatchNorm2d(self.out_channels))
            feats.add_module('final_relu', nn.ReLU(True))
        self.features = feats

    def to_bottleneck(self) -> 'ResNet':
        self._change_block_type('bottleneck')
        return self

    def to_preact_bottleneck(self) -> 'ResNet':
        self._change_block_type('preact_bottleneck')
        return self

    def to_preact(self) -> 'ResNet':
        self._change_block_type('preact_basic')
        return self

    def to_resnext(self) -> 'ResNet':
        self._change_block_type('resnext')
        return self

    def to_preact_resnext(self) -> 'ResNet':
        self._change_block_type('preact_resnext')
        return self

    def to_wide(self) -> 'ResNet':
        self._change_block_type('wide')
        return self

    def to_preact_wide(self) -> 'ResNet':
        self._change_block_type('preact_wide')
        return self

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.classifier(self.features(x))

    def add_se(self) -> 'ResNet':
        for m in self.features:
            if isinstance(m, BLOCKS):
                m.use_se()
        return self

    def set_input_specs(self,
                        input_size: int = 224,
                        in_channels: int = 3) -> 'ResNet':
        assert isinstance(self.features.input, ResNetInput)
        self.features.input.set_input_specs(input_size=input_size,
                                            in_channels=in_channels)
        return self


#
# CIFAR
#


def resnet20_cifar() -> ResNet:
    return ResNet([
        '16:1', '16:1', '16:1', '32:2', '32:1', '32:1', '64:2', '64:1', '64:1'
    ], 10).set_input_specs(input_size=32)


def preact_resnet20_cifar() -> ResNet:
    return resnet20_cifar().to_preact()


#
# ResNets
#


def resnet18(num_classes: int) -> ResNet:
    return ResNet(
        ['64:1', '64:1', '128:2', '128:1', '256:2', '256:1', '512:2', '512:1'],
        num_classes)


def resnet34(num_classes: int) -> ResNet:
    return ResNet(['64:1'] * 3 + ['128:2'] + ['128:1'] * 3 + ['256:2']
                  + ['256:1'] * 5 + ['512:2', '512:1', '512:1'], num_classes)


def resnet50(num_classes: int) -> ResNet:
    net = ResNet(['256:1'] * 3 + ['512:2'] + ['512:1'] * 3 + ['1024:2']
                 + ['1024:1'] * 5 + ['2048:2', '2048:1', '2048:1'], num_classes)
    net.to_bottleneck()
    return net


def resnet101(num_classes: int) -> ResNet:
    net = ResNet(['256:1'] * 3 + ['512:2'] + ['512:1'] * 3 + ['1024:2']
                 + ['1024:1'] * 22 + ['2048:2', '2048:1', '2048:1'], num_classes)
    net.to_bottleneck()
    return net


def resnet152(num_classes: int) -> ResNet:
    net = ResNet(['256:1'] * 3 + ['512:2'] + ['512:1'] * 7 + ['1024:2']
                 + ['1024:1'] * 35 + ['2048:2', '2048:1', '2048:1'], num_classes)
    net.to_bottleneck()
    return net


#
# Preact ResNet
#


def preact_resnet18(num_classes: int) -> ResNet:
    return resnet18(num_classes).to_preact()


def preact_resnet34(num_classes: int) -> ResNet:
    return resnet34(num_classes).to_preact()


def preact_resnet50(num_classes: int) -> ResNet:
    return resnet50(num_classes).to_preact_bottleneck()


def preact_resnet101(num_classes: int) -> ResNet:
    return resnet101(num_classes).to_preact_bottleneck()


def preact_resnet152(num_classes: int) -> ResNet:
    return resnet152(num_classes).to_preact_bottleneck()


#
# ResNeXt
#


def resnext50_32x4d(num_classes: int) -> ResNet:
    return resnet50(num_classes).to_resnext()


def resnext101_32x4d(num_classes: int) -> ResNet:
    return resnet101(num_classes).to_resnext()


def resnext152_32x4d(num_classes: int) -> ResNet:
    return resnet152(num_classes).to_resnext()


#
# Preact ResNeXt
#


def preact_resnext50_32x4d(num_classes: int) -> ResNet:
    return resnet50(num_classes).to_preact_resnext()


def preact_resnext101_32x4d(num_classes: int) -> ResNet:
    return resnet101(num_classes).to_preact_resnext()


def preact_resnext152_32x4d(num_classes: int) -> ResNet:
    return resnet152(num_classes).to_preact_resnext()


#
# Wide
#

def preact_wide_resnet50(num_classes: int) -> ResNet:
    return resnet50(num_classes).to_preact_wide()


def preact_wide_resnet101(num_classes: int) -> ResNet:
    return resnet101(num_classes).to_preact_wide()

#
# Wide
#


def wide_resnet50(num_classes: int) -> ResNet:
    return resnet50(num_classes).to_wide()


def wide_resnet101(num_classes: int) -> ResNet:
    return resnet101(num_classes).to_wide()



from collections import OrderedDict
import torch
import torch.nn as nn

from torchelie.utils import experimental
import torchelie.nn as tnn
from torchelie.utils import kaiming


class ClassificationHead(tnn.CondSeq):
    """
    A one layer classification head, turning activations / features into class
    log probabilities.

    It initially contains an avgpool-flatten-linear architecture.

    Args:
        in_channels (int): the number of features in the last layer of the
            feature extractor
        num_classes (int): the number of output classes
    """
    linear1: nn.Linear

    def __init__(self, in_channels: int, num_classes: int) -> None:
        super(ClassificationHead, self).__init__()
        self.in_channels = in_channels
        self.num_classes = num_classes

        self.to_resnet_style()

    def to_resnet_style(self) -> 'ClassificationHead':
        """
        Set the classifier architecture to avgpool-flatten-linear.
        """
        self._modules = OrderedDict([
            ('pool', nn.AdaptiveAvgPool2d(1)),
            ('reshape', tnn.Reshape(self.in_channels)),
            ('linear1', kaiming(nn.Linear(self.in_channels,
                                          self.num_classes))),
        ])

        return self

    def to_two_layers(self, hidden_channels: int) -> 'ClassificationHead':
        """
        Set the classifier architecture to avgpool-flatten-linear1-relu-linear2.
        """
        self._modules = OrderedDict([
            ('pool', nn.AdaptiveAvgPool2d(1)),
            ('reshape', tnn.Reshape(self.in_channels)),
            ('linear1', kaiming(nn.Linear(self.in_channels, hidden_channels))),
            ('relu1', nn.ReLU(True)),
            ('linear2', kaiming(nn.Linear(hidden_channels, self.num_classes))),
        ])
        return self

    def to_vgg_style(self, hidden_channels: int) -> 'ClassificationHead':
        """
        Set the classifier architecture to
        avgpool-flatten-linear1-relu-dropout-linear2-relu-dropout-linear3, like
        initially done with VGG.
        """
        self._modules = OrderedDict([
            ('pool', nn.AdaptiveAvgPool2d(7)),
            ('reshape', tnn.Reshape(7 * 7 * self.in_channels)),
            ('linear1',
             kaiming(nn.Linear(7 * 7 * self.in_channels, hidden_channels))),
            ('relu1', nn.ReLU(True)),
            ('dropout1', nn.Dropout(0.5)),
            ('linear2', kaiming(nn.Linear(hidden_channels, hidden_channels))),
            ('relu2', nn.ReLU(True)),
            ('dropout2', nn.Dropout(0.5)),
            ('linear3', kaiming(nn.Linear(hidden_channels, self.num_classes))),
        ])

        return self

    def to_convolutional(self) -> 'ClassificationHead':
        """
        Remove pooling and flattening operations, convert linears to conv1x1
        """
        del self.reshape
        del self.pool

        def _do(m):
            if isinstance(m, nn.Linear):
                return tnn.Conv1x1(m.in_features, m.out_features)
            return m
        tnn.utils.edit_model(self, _do)
        return self

    def leaky(self) -> 'ClassificationHead':
        """
        Make relus leaky
        """
        tnn.utils.make_leaky(self)
        return self

    def set_num_classes(self, classes: int) -> 'ClassificationHead':
        """
        change the number of output classes
        """
        self.num_classes = classes
        old = self[-1]
        self[-1] = kaiming(nn.Linear(old.in_features, self.num_classes))
        return self

    def remove_pool(self, spatial_size: int) -> 'ClassificationHead':
        """
        remove the pooling operation
        """
        self.set_pool_size(spatial_size)
        del self.pool
        return self

    def set_pool_size(self, size: int) -> 'ClassificationHead':
        """
        Average pool to spatial size :code:`size` rather than 1. Recreate the
        first Linear to accomodate the change.
        """
        self.pool = nn.AdaptiveAvgPool2d(size)
        self.reshape = tnn.Reshape(size * size * self.in_channels)
        self.linear1 = kaiming(
            nn.Linear(size * size * self.in_channels,
                      self.linear1.out_features))
        return self

    @experimental
    def rm_dropout(self) -> 'ClassificationHead':
        """
        Remove the dropout layers if any.
        """
        if hasattr(self, 'dropout1'):
            del self.dropout1
        if hasattr(self, 'dropout2'):
            del self.dropout2
        return self


@experimental
class ProjectionDiscr(nn.Module):
    """
    A classification head for conditional GANs discriminators using a
    `projection discriminator <https://arxiv.org/abs/1802.05637>`_ .

    Args:
        feat_extractor (nn.Module): a feature extraction model
        feature_size (int): the number of features in the last layer of the
            feature extractor
        num_classes (int): the number of output classes
    """
    def __init__(self, in_channels: int, num_classes: int) -> None:
        super().__init__()
        self.head = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            tnn.Reshape(in_channels),
        )
        self.emb = nn.Embedding(num_classes, in_channels)
        self.discr = nn.Linear(in_channels, 1)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Forward pass

        Args:
            x: argument for `feat_extractor`
            y: class label
        """
        feats = self.head(x)
        y_emb = self.emb(y)
        return self.discr(feats) + torch.mm(y_emb, feats.t())



import torch
import torch.nn as nn
from .functional import focal_loss


class FocalLoss(nn.Module):
    """
    The focal loss

    https://arxiv.org/abs/1708.02002

    See :func:`torchelie.loss.focal_loss` for details.
    """
    def __init__(self, gamma: float = 0):
        super(FocalLoss, self).__init__()
        self.gamma = gamma

    def forward(self, input: torch.Tensor,
                target: torch.Tensor) -> torch.Tensor:
        return focal_loss(input, target, self.gamma)



import torch
import torch.nn as nn

import torchelie.loss.gan
from .functional import ortho, total_variation, continuous_cross_entropy
from .functional import focal_loss

from .perceptualloss import PerceptualLoss
from .neuralstyleloss import NeuralStyleLoss
from .deepdreamloss import DeepDreamLoss
from .focal import FocalLoss
from .bitempered import tempered_cross_entropy, TemperedCrossEntropyLoss
from .bitempered import tempered_softmax, tempered_nll_loss
from .bitempered import tempered_log_softmax


class OrthoLoss(nn.Module):
    """
    Orthogonal loss

    See :func:`torchelie.loss.ortho` for details.
    """
    def forward(self, w):
        return ortho(w)


class TotalVariationLoss(nn.Module):
    """
    Total Variation loss

    See :func:`torchelie.loss.total_variation` for details.
    """
    def forward(self, x):
        return total_variation(x)


class ContinuousCEWithLogits(nn.Module):
    """
    Cross Entropy loss accepting continuous target values

    See :func:`torchelie.loss.continuous_cross_entropy` for details.
    """
    def forward(self, pred, soft_targets):
        return continuous_cross_entropy(pred, soft_targets)


def binary_hinge(x, y):
    p = y * torch.clamp(1 - x, min=0) + (1 - y) * torch.clamp(1 + x, min=0)
    return p.mean()



import torch
import torch.nn as nn
import torchelie.utils as tu
import torchelie.callbacks as tcb
from torchelie.recipes.recipebase import Recipe
from typing import Optional, Iterable, Any


@tu.experimental
def GANRecipe(G: nn.Module,
              D: nn.Module,
              G_fun,
              D_fun,
              test_fun,
              loader: Iterable[Any],
              *,
              visdom_env: Optional[str] = 'main',
              checkpoint: Optional[str] = 'model',
              test_every: int = 1000,
              log_every: int = 10,
              g_every: int = 1) -> Recipe:
    def D_wrap(batch):
        tu.freeze(G)
        tu.unfreeze(D)

        return D_fun(batch)

    def G_wrap(batch):
        tu.freeze(D)
        tu.unfreeze(G)

        return G_fun(batch)

    def test_wrap(batch):
        tu.freeze(G)
        tu.freeze(D)
        D.eval()
        G.eval()

        with torch.no_grad():
            out = test_fun(batch)

        D.train()
        G.train()
        return out

    class NoLim:
        def __init__(self):
            self.i = iter(loader)
            self.did_send = False

        def __iter__(self):
            return self

        def __next__(self):
            if self.did_send:
                self.did_send = False
                raise StopIteration
            self.did_send = True
            try:
                return next(self.i)
            except:
                self.i = iter(loader)
                return next(self.i)

    D_loop = Recipe(D_wrap, loader)
    D_loop.register('G', G)
    D_loop.register('D', D)
    G_loop = Recipe(G_wrap, NoLim())
    D_loop.G_loop = G_loop
    D_loop.register('G_loop', G_loop)

    test_loop = Recipe(test_wrap, NoLim())
    D_loop.test_loop = test_loop
    D_loop.register('test_loop', test_loop)

    def G_test(state):
        G_loop.callbacks.update_state({
            'epoch': state['epoch'],
            'iters': state['iters'],
            'epoch_batch': state['epoch_batch']
        })

    def prepare_test(state):
        test_loop.callbacks.update_state({
            'epoch': state['epoch'],
            'iters': state['iters'],
            'epoch_batch': state['epoch_batch']
        })

    D_loop.callbacks.add_prologues([tcb.Counter()])

    D_loop.callbacks.add_epilogues([
        tcb.Log('imgs', 'G_imgs'),
        tcb.CallRecipe(G_loop, g_every, init_fun=G_test, prefix='G'),
        tcb.VisdomLogger(visdom_env=visdom_env, log_every=log_every),
        tcb.StdoutLogger(log_every=log_every),
        tcb.CallRecipe(test_loop,
                       test_every,
                       init_fun=prepare_test,
                       prefix='Test'),
    ])

    G_loop.callbacks.add_epilogues([
        tcb.WindowedMetricAvg('G_loss'),
        tcb.VisdomLogger(visdom_env=visdom_env,
                         log_every=log_every,
                         post_epoch_ends=False)
    ])

    if checkpoint is not None:
        test_loop.callbacks.add_epilogues([
            tcb.Checkpoint(checkpoint + '/ckpt_{iters}.pth', D_loop),
            tcb.VisdomLogger(visdom_env=visdom_env),
        ])

    return D_loop



import torch
import torch.nn.functional as F
import torch.nn as nn

import torchelie.nn as tnn


class DeepDreamLoss(nn.Module):
    """
    The Deep Dream loss

    Args:
        model (nn.Module): a pretrained network on which to compute the
            activations
        dream_layer (str): the name of the layer on which the activations are
            to be maximized
        max_reduction (int): the maximum factor of reduction of the image, for
            multiscale generation
    """
    def __init__(self,
                 model: nn.Module,
                 dream_layer: str,
                 max_reduction: int = 3) -> None:
        super(DeepDreamLoss, self).__init__()
        self.dream_layer = dream_layer
        self.octaves = max_reduction
        model = model.eval()
        self.net = tnn.WithSavedActivations(model, names=[self.dream_layer])
        self.i = 0

    def get_acts_(self, img: torch.Tensor, detach: bool) -> torch.Tensor:
        octave = (self.i % (self.octaves * 2)) / 2 + 1
        this_sz_img = F.interpolate(img, scale_factor=1 / octave)
        _, activations = self.net(this_sz_img, detach=detach)
        return activations[self.dream_layer]

    def forward(self, input_img: torch.Tensor) -> torch.Tensor:
        """
        Compute the Deep Dream loss on `input_img`
        """
        dream = self.get_acts_(input_img, detach=False)
        self.i += 1

        dream_loss = -dream.pow(2).sum()

        return dream_loss



import torch

from torchelie.loss import *
import torchelie.loss.gan as gan
import torchelie.loss.functional as tlf


def test_bitempered():
    x = torch.randn(3, 5)
    y = torch.arange(3)

    tsm = tempered_softmax(x, 1)
    sm = torch.nn.functional.softmax(x, dim=1)
    assert sm.allclose(tsm)

    tsm = tempered_log_softmax(x, 1)
    sm = torch.nn.functional.log_softmax(x, dim=1)
    assert sm.allclose(tsm)

    tnll = tempered_nll_loss(sm, y, 1, 1)
    nll = torch.nn.functional.nll_loss(sm, y)
    assert nll.allclose(tnll)

    temp_loss = tempered_cross_entropy(x, y, 1, 1)
    ce_loss = torch.nn.functional.cross_entropy(x, y)
    assert ce_loss.allclose(temp_loss)

    tce = TemperedCrossEntropyLoss(1, 1)
    assert ce_loss.allclose(tce(x, y))


def test_deepdream():
    m = nn.Sequential(nn.Conv2d(1, 1, 3))
    dd = DeepDreamLoss(m, '0', max_reduction=1)
    dd(m(torch.randn(1, 1, 10, 10)))


def test_focal():
    y = (torch.randn(10, 1) < 0).float()
    x = torch.randn(10, 1)

    foc = FocalLoss()
    fl = foc(x, y)
    loss = torch.nn.functional.binary_cross_entropy_with_logits(x, y)
    assert torch.allclose(fl, loss)

    y = torch.randint(4, (10,))
    x = torch.randn(10, 5)

    fl = foc(x, y)
    loss = torch.nn.functional.cross_entropy(x, y)
    assert torch.allclose(fl, loss)

    focal_loss(torch.randn(10, 5), torch.randint(4, (10,)))


def test_funcs():
    f = OrthoLoss()
    f(torch.randn(10, 10))
    ortho(torch.randn(10, 10))

    f = TotalVariationLoss()
    f(torch.randn(1, 1, 10, 10))
    total_variation(torch.randn(1, 1, 10, 10))

    f = ContinuousCEWithLogits()
    continuous_cross_entropy(torch.randn(10, 5),
                             torch.nn.functional.softmax(torch.randn(10, 5), 1))
    f(torch.randn(10, 5), torch.nn.functional.softmax(torch.randn(10, 5), 1))


def test_neural_style():
    ns = NeuralStyleLoss()
    ns.set_content(torch.randn(3, 128, 128))
    ns.set_style(torch.randn(3, 128, 128), 1)
    ns(torch.randn(1, 3, 128, 128))


def test_perceptual():
    pl = PerceptualLoss(['conv1_1'], rescale=True)
    pl(torch.randn(1, 3, 64, 64), torch.randn(1, 3, 64, 64))


def test_gan():
    x = torch.randn(5, 5)

    gan.hinge.real(x)
    gan.hinge.fake(x)
    gan.hinge.generated(x)

    gan.standard.real(x)
    gan.standard.fake(x)
    gan.standard.generated(x)



import torch
import torch.nn as nn
from torchelie.utils import kaiming_gain, kaiming, fast_zero_grad
from torch.nn import Module
from collections import OrderedDict
from torch.nn.parameter import Parameter
from typing import Any, TypeVar, Callable, Tuple


class WeightLambda:
    """
    Apply a lambda function as a hook to the weight matrix of a layer before
    a forward pass.

    Don't use it directly, use the functions :code:`weight_lambda()` and
    :code:`remove_weight_lambda()` instead.

    Args:
        hook_name (str): an identifier for that WeightLambda hook, such as
            'l2normalize', 'weight_norm', etc.
        name (str): the name of the module's parameter to apply the hook on
        function (Callable): a function of the form
            :code:`(torch.Tensor) -> torch.Tensor` that takes applies the
            desired computation to the module's parameter.
    """
    name: str

    def __init__(self, hook_name: str, name: str, function) -> None:
        self.name = name
        self.hook_name = hook_name
        self.fun = function

    @staticmethod
    def apply(module, hook_name: str, name: str, function) -> 'WeightLambda':
        fn = WeightLambda(hook_name, name, function)

        weight = getattr(module, name)
        # remove w from parameter list
        del module._parameters[name]

        module.register_parameter(name + '_g', Parameter(weight.data))
        setattr(module, name, fn.fun(getattr(module, name + '_g')))

        # recompute weight before every forward()
        module.register_forward_pre_hook(fn)

        return fn

    def remove(self, module: Module) -> None:
        weight = self.fun(module)
        delattr(module, self.name)
        del module._parameters[self.name + '_g']
        setattr(module, self.name, Parameter(weight.data))

    def __call__(self, module: Module, inputs: Any) -> None:
        setattr(module, self.name, self.fun(getattr(module, self.name + '_g')))


def weight_lambda(
    module: Module,
    hook_name: str,
    function,
    name: str = 'weight',
) -> Module:
    """
    Apply :code:`function()` to :code:`getattr(module, name)` on each forward
    pass.

    Allows to implement things such as weight normalization, or equalized
    learning rate weight scaling.

    Args:
        module (nn.Module): the module to hook on
        hook_name (str): an identifier for that WeightLambda hook, such as
            'l2normalize', 'weight_norm', etc.
        function (Callable): a function of the form
            :code:`(torch.Tensor) -> torch.Tensor` that takes applies the
            desired computation to the module's parameter.
        name (str): the name of the module's parameter to apply the hook on.
            Default: 'weight'.

    Returns:
        the module with the hook
    """
    WeightLambda.apply(module, hook_name, name, function)
    return module


def remove_weight_lambda(module: Module,
                         hook_name: str,
                         name: str = 'weight') -> Module:
    """
    Remove the hook :code:`hook_name` applied on member :code:`name` of
    :code:`module`.

    Args:
        module (nn.Module): the module on which the hook has to be removed
        hook_name (str): an identifier for that WeightLambda hook, such as
            'l2normalize', 'weight_norm', etc.
        name (str): the name of the module's parameter the hook is applied on.
            Default: 'weight'.
    """
    for k, hook in module._forward_pre_hooks.items():
        if (isinstance(hook, WeightLambda) and hook.name == name
                and hook.hook_name == hook_name):
            hook.remove(module)
            del module._forward_pre_hooks[k]
            return module

    raise ValueError("weight_lambda of '{}' not found in {}".format(
        hook_name, module))


def weight_scale(module: Module,
                 name: str = 'weight',
                 scale: float = 0) -> Module:
    """
    Multiply :code:`getattr(module, name)` by :code:`scale` on forward pass
    as a hook. Used to implement equalized LR for StyleGAN
    """
    return weight_lambda(module, 'scale', lambda w: w * scale, name)


def remove_weight_scale(module: Module, name: str = 'weight') -> Module:
    """
    Remove a weight_scale hook previously applied on
    :code:`getattr(module, name)`.
    """
    return remove_weight_lambda(module, 'scale', name)


@torch.no_grad()
def remove_batchnorm(m: nn.Sequential) -> None:
    """
    Remove BatchNorm in Sequentials / CondSeqs in a smart way, restoring biases
    in the preceding layer.
    """
    ms = list(m._modules.items())

    # transfer biases from BN to previous conv / Linear / Whatever
    for (name1, mod1), (name2, mod2) in zip(ms[:-1], ms[1:]):
        if isinstance(mod2, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)):
            if mod1.bias is not None:
                continue

            if mod2.bias is not None:
                with torch.no_grad():
                    mod1.bias = mod2.bias
            else:
                out_ch = len(mod2.running_mean)
                with torch.no_grad():
                    mod1.bias = nn.Parameter(torch.zeros(out_ch))
    # remove bn
    for name, mod in ms:
        if isinstance(mod, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)):
            delattr(m, name)


T_Module = TypeVar('T_Module', bound=nn.Module)


@torch.no_grad()
def edit_model(m: T_Module, f: Callable[[nn.Module], nn.Module]) -> nn.Module:
    """
    Allow to edit any part of a model by recursively editing its modules.

    For instance, in order to delete all dropout layers and change relus into
    leakyrelus

    ::

        def make_leaky_no_dropout(m):
            if isinstance(m, nn.ReLU):
                return nn.LeakyReLU(inplace=True)
            if isinstance(m, nn.Dropout2d):
                return nn.Identity()
            return m
        model = edit_model(model, make_leaky_no_dropout)

    Args:
        m (nn.Module): the model to edit
        f (Callabble: nn.Module -> nn.Module): a mapping function applied to
            all modules and submodules

    Returns:
        The edited model.
    """
    for name, mod in m._modules.items():
        m._modules[name] = edit_model(mod, f)
        m._modules[name] = f(mod)
    return f(m)


@torch.no_grad()
def weight_norm_and_equal_lr(m: T_Module,
                             leak: float = 0.,
                             mode: str = 'fan_in',
                             init_gain: float = 1.,
                             lr_gain: float = 1.,
                             name: str = 'weight') -> T_Module:
    """
    Set weight norm and equalized learning rate like demodulated convs in
    StyleGAN2 for module m.

    The weight matrix is initialized for a leaky relu nonlinearity of slope a.
    An extra gain can be specified, as well as a differential learning rate
    multiplier.

    See StyleGAN2 paper for more info.

    """
    kai_gain = kaiming_gain(m, a=leak, mode=mode)
    gain = kai_gain * init_gain * lr_gain

    def do_it(w):
        w = w * gain
        shape = w.shape
        w_flat = w.view(w.shape[0], -1)
        w_flat = w_flat / w_flat.norm(dim=1, keepdim=True)
        return w_flat.view(*shape)

    m.weight.data.normal_(0., 1. / lr_gain)
    weight_lambda(m, 'norm_equal_lr', do_it, name=name)
    return m


@torch.no_grad()
def remove_weight_norm_and_equal_lr(module: Module,
                                    name: str = 'weight') -> Module:
    """
    Remove a weight_norm_and_equal_lr hook previously applied on
    :code:`getattr(module, name)`.
    """
    return remove_weight_lambda(module, 'norm_equal_lr', name)


@torch.no_grad()
def insert_after(base: nn.Sequential, key: str, new: nn.Module,
                 name: str) -> nn.Sequential:
    """
    Insert module :code:`new` with name :code:`name` after element :code:`key`
    in sequential :code:`base` and return the new sequence.
    """
    modules_list = list(base._modules.items())
    found = -1
    for i, (nm, m) in enumerate(modules_list):
        if nm == key:
            found = i
            break
    assert found != -1
    modules_list.insert(found + 1, (name, new))
    base._modules = OrderedDict(modules_list)
    return base


@torch.no_grad()
def insert_before(base: nn.Sequential, key: str, new: nn.Module,
                  name: str) -> nn.Sequential:
    """
    Insert module :code:`new` with name :code:`name` before element :code:`key`
    in sequential :code:`base` and return the new sequence.
    """
    modules_list = list(base._modules.items())
    found = -1
    for i, (nm, m) in enumerate(modules_list):
        if nm == key:
            found = i
            break
    assert found != -1
    modules_list.insert(found, (name, new))
    base._modules = OrderedDict(modules_list)
    return base


@torch.no_grad()
def make_leaky(net: nn.Module) -> nn.Module:
    """
    Change all relus into leaky relus for modules and submodules of net.
    """

    def do_it(m: nn.Module) -> nn.Module:
        if isinstance(m, nn.ReLU):
            return nn.LeakyReLU(0.2, m.inplace)
        return m

    return edit_model(net, do_it)


@torch.no_grad()
def net_to_equal_lr(net: nn.Module,
                    leak: float = 0.,
                    mode: str = 'fan_in') -> T_Module:
    """
    Set all Conv2d, ConvTransposed2d and Linear of :code:`net` to equalized
    learning rate, initialized with :func:`torchelie.utils.kaiming` and
    :code:`dynamic=True`.

    Returns:
        :code:`net`.
    """
    for m in net.modules():
        if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d, nn.Linear)):
            kaiming(m, a=leak, mode=mode, dynamic=True)
    return net


@torch.no_grad()
def net_to_weight_norm_and_equal_lr(net: T_Module,
                                    leak: float = 0.,
                                    mode: str = 'fan_in',
                                    init_gain: float = 1.,
                                    lr_gain: float = 1.) -> T_Module:
    """
    Set all Conv2d, ConvTransposed2d and Linear of :code:`net` to equalized
    learning rate and weight normalization, initialized with
    :func:`torchelie.utils.weight_norm_and_equal_lr`.

    Returns:
        :code:`net`.
    """
    for m in net.modules():
        if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d, nn.Linear)):
            weight_norm_and_equal_lr(m,
                                     leak=leak,
                                     mode=mode,
                                     init_gain=init_gain,
                                     lr_gain=lr_gain)
    return net


def receptive_field_for(net: nn.Module,
                        input_size: Tuple[int, int, int]) -> Tuple[int, int]:
    """
    Compute the receptive field of :code:`net` using a backward pass.

    .. warning::
        Make sure the net does not have any normalization operation: mean /
        std operations impact the whole volume and make the calculation
        invalid. (BatchNorm is okay in eval mode)

    Args:
        input_size (Tuple[int, int, int]): a (C, H, W) volume size that the
            model can accept. Note that the receptive field calculation is
            bounded by this size.

    Returns:
        height and width receptive fields.
    """
    device = next(iter(net.parameters())).device
    input = torch.randn(1, *input_size, requires_grad=True, device=device)
    net.eval()
    out = net(input)
    net.train()
    if out.dim() == 2:
        out.mean().backward()
    elif out.dim() == 4:
        out[:, :, out.shape[2] // 2, out.shape[3] // 2].mean().backward()
    fast_zero_grad(net)
    has_grad = torch.nonzero(input.grad[0, 0])
    recep_field = has_grad.max(dim=0).values - has_grad.min(dim=0).values + 1
    return recep_field[0].item(), recep_field[1].item()
